import time
import numpy as np
import os.path as pa
from astropy.io import fits
from sfft.sfftcore.SFFTSubtract import GeneralSFFTSubtract
from sfft.sfftcore.SFFTConfigure import SingleSFFTConfigure
# version: Apr 22, 2025

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6"

class Customized_Packet:
    @staticmethod
    def CP(FITS_REF, FITS_SCI, FITS_mREF, FITS_mSCI, ForceConv, GKerHW, \
        FITS_DIFF=None, FITS_Solution=None, KerPolyOrder=2, BGPolyOrder=2, ConstPhotRatio=True, \
        BACKEND_4SUBTRACT='Cupy', CUDA_DEVICE_4SUBTRACT='0', NUM_CPU_THREADS_4SUBTRACT=8, NUMBA_CACHE=True, VERBOSE_LEVEL=2):
        
        """
        * Parameters for Customized SFFT
        # @ Customized: Skip built-in preprocessing (automatic image-masking) by a customized masked image-pair.
        #
        # ----------------------------- Computing Enviornment --------------------------------- #

        -BACKEND_4SUBTRACT ['Cupy']         # 'Cupy' or 'Numpy'.
                                            # Cupy backend require GPU(s) that is capable of performing double-precision calculations,
                                            # while Numpy backend is a pure CPU-based backend for sfft subtraction.
                                            # NOTE: 'Pycuda' backend is no longer supported since sfft v1.4.0.

        -CUDA_DEVICE_4SUBTRACT ['0']        # it specifies certain GPU device (index) to conduct the subtraction task.
                                            # the GPU devices are usually numbered 0 to N-1 (you may use command nvidia-smi to check).
                                            # this argument becomes trivial for Numpy backend.

        -NUM_CPU_THREADS_4SUBTRACT [8]      # it specifies the number of CPU threads used for sfft subtraction in Numpy backend.
                                            # SFFT in Numpy backend has been implemented with pyFFTW and numba, 
                                            # that allow for parallel computing on CPUs. Of course, the Numpy 
                                            # backend is generally much slower than GPU backends.
        
        -NUMBA_CACHE [True]                # it determines whether to cache the compiled functions in numba.
                                           # if True, the compiled functions will be cached in the local directory __pycache__ or a specified NUMBA_CACHE_DIR.
                                           # it can save the compilation time when the functions are called multiple times with same configuration for different tasks.

        # ----------------------------- SFFT Subtraction --------------------------------- #

        -ForceConv []                       # it determines which image will be convolved, can be 'REF' or 'SCI'.
                                            # here ForceConv CANNOT be 'AUTO'!

        -GKerHW []                          # the given kernel half-width. 

        -KerPolyOrder [2]                   # Polynomial degree of kernel spatial variation.

        -BGPolyOrder [2]                    # Polynomial degree of background spatial variation.

        -ConstPhotRatio [True]              # Constant photometric ratio between images ? can be True or False
                                            # ConstPhotRatio = True: the sum of convolution kernel is restricted to be a 
                                            #                constant across the field. 
                                            # ConstPhotRatio = False: the flux scaling between images is modeled by a 
                                            #                polynomial with degree -KerPolyOrder.

        # ----------------------------- Input & Output --------------------------------- #

        -FITS_REF []                        # File path of input reference image.

        -FITS_SCI []                        # File path of input science image.

        -FITS_mREF []                       # File path of input masked reference image (NaN-free).

        -FITS_mSCI []                       # File path of input masked science image (NaN-free).

        -FITS_DIFF [None]                   # File path of output difference image.

        -FITS_Solution [None]               # File path of the solution of the linear system.
                                            # it is an array of (..., a_ijab, ... b_pq, ...).

        # ----------------------------- Miscellaneous --------------------------------- #
        
        -VERBOSE_LEVEL [2]                  # The level of verbosity, can be [0, 1, 2]
                                            # 0/1/2: QUIET/NORMAL/FULL mode

        # Important Notice:
        #
        # a): if reference is convolved in SFFT (-ForceConv='REF'), then DIFF = SCI - Convolved_REF.
        #     [the difference image is expected to have PSF & flux zero-point consistent with science image]
        #
        # b): if science is convolved in SFFT (-ForceConv='SCI'), then DIFF = Convolved_SCI - REF
        #     [the difference image is expected to have PSF & flux zero-point consistent with reference image]
        #
        # Remarks: this convention is to guarantee that transients emerge on science image always 
        #          show a positive signal on difference images.

        """

        # * Read input images
        PixA_REF = fits.getdata(FITS_REF, ext=0).T
        PixA_SCI = fits.getdata(FITS_SCI, ext=0).T
        PixA_mREF = fits.getdata(FITS_mREF, ext=0).T
        PixA_mSCI = fits.getdata(FITS_mSCI, ext=0).T

        if not PixA_REF.flags['C_CONTIGUOUS']:
            PixA_REF = np.ascontiguousarray(PixA_REF, np.float64)
        else: PixA_REF = PixA_REF.astype(np.float64)

        if not PixA_SCI.flags['C_CONTIGUOUS']:
            PixA_SCI = np.ascontiguousarray(PixA_SCI, np.float64)
        else: PixA_SCI = PixA_SCI.astype(np.float64)

        if not PixA_mREF.flags['C_CONTIGUOUS']:
            PixA_mREF = np.ascontiguousarray(PixA_mREF, np.float64)
        else: PixA_mREF = PixA_mREF.astype(np.float64)

        if not PixA_mSCI.flags['C_CONTIGUOUS']:
            PixA_mSCI = np.ascontiguousarray(PixA_mSCI, np.float64)
        else: PixA_mSCI = PixA_mSCI.astype(np.float64)

        NaNmask_U = None
        NaNmask_REF = np.isnan(PixA_REF)
        NaNmask_SCI = np.isnan(PixA_SCI)
        if NaNmask_REF.any() or NaNmask_SCI.any():
            NaNmask_U = np.logical_or(NaNmask_REF, NaNmask_SCI)

        assert np.sum(np.isnan(PixA_mREF)) == 0
        assert np.sum(np.isnan(PixA_mSCI)) == 0
        
        assert ForceConv in ['REF', 'SCI']
        ConvdSide = ForceConv
        KerHW = GKerHW

        # * Choose GPU device for Cupy backend
        if BACKEND_4SUBTRACT == 'Cupy':
            import cupy as cp
            device = cp.cuda.Device(int(CUDA_DEVICE_4SUBTRACT))
            device.use()

        # * Compile Functions in SFFT Subtraction
        if VERBOSE_LEVEL in [0, 1, 2]:
            print('MeLOn CheckPoint: TRIGGER Function Compilations of SFFT-SUBTRACTION!')

        Tcomp_start = time.time()
        SFFTConfig = SingleSFFTConfigure.SSC(NX=PixA_REF.shape[0], NY=PixA_REF.shape[1], KerHW=KerHW, \
            KerPolyOrder=KerPolyOrder, BGPolyOrder=BGPolyOrder, ConstPhotRatio=ConstPhotRatio, \
            BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, \
            NUMBA_CACHE=NUMBA_CACHE, VERBOSE_LEVEL=VERBOSE_LEVEL)

        if VERBOSE_LEVEL in [1, 2]:
            _message = 'Function Compilations of SFFT-SUBTRACTION TAKES [%.3f s]' %(time.time() - Tcomp_start)
            print('\nMeLOn Report: %s' %_message)

        # * Perform SFFT Subtraction
        if ConvdSide == 'REF':
            PixA_mI, PixA_mJ = PixA_mREF, PixA_mSCI
            if NaNmask_U is not None:
                PixA_I, PixA_J = PixA_REF.copy(), PixA_SCI.copy()
                PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
            else: PixA_I, PixA_J = PixA_REF, PixA_SCI

        if ConvdSide == 'SCI':
            PixA_mI, PixA_mJ = PixA_mSCI, PixA_mREF
            if NaNmask_U is not None:
                PixA_I, PixA_J = PixA_SCI.copy(), PixA_REF.copy()
                PixA_I[NaNmask_U] = PixA_mI[NaNmask_U]
                PixA_J[NaNmask_U] = PixA_mJ[NaNmask_U]
            else: PixA_I, PixA_J = PixA_SCI, PixA_REF
        
        if VERBOSE_LEVEL in [0, 1, 2]:
            print('MeLOn CheckPoint: TRIGGER SFFT-SUBTRACTION!')

        Tsub_start = time.time()
        _tmp = GeneralSFFTSubtract.GSS(PixA_I=PixA_I, PixA_J=PixA_J, PixA_mI=PixA_mI, PixA_mJ=PixA_mJ, \
            SFFTConfig=SFFTConfig, ContamMask_I=None, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
            NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, VERBOSE_LEVEL=VERBOSE_LEVEL)

        Solution, PixA_DIFF = _tmp[:2]
        if VERBOSE_LEVEL in [1, 2]:
            _message = 'SFFT-SUBTRACTION TAKES [%.3f s]' %(time.time() - Tsub_start)
            print('\nMeLOn Report: %s' %_message)
        
        # * Modifications on the difference image
        #   a) when REF is convolved, DIFF = SCI - Conv(REF)
        #      PSF(DIFF) is coincident with PSF(SCI), transients on SCI are positive signal in DIFF.
        #   b) when SCI is convolved, DIFF = Conv(SCI) - REF
        #      PSF(DIFF) is coincident with PSF(REF), transients on SCI are still positive signal in DIFF.

        if NaNmask_U is not None:
            # ** Mask Union-NaN region
            PixA_DIFF[NaNmask_U] = np.nan
        if ConvdSide == 'SCI': 
            # ** Flip difference when science is convolved
            PixA_DIFF = -PixA_DIFF
        
        # * Save difference image
        if FITS_DIFF is not None:
            _hdl = fits.open(FITS_SCI)
            _hdl[0].data[:, :] = PixA_DIFF.T
            _hdl[0].header['NAME_REF'] = (pa.basename(FITS_REF), 'MeLOn: SFFT')
            _hdl[0].header['NAME_SCI'] = (pa.basename(FITS_SCI), 'MeLOn: SFFT')
            _hdl[0].header['KERORDER'] = (KerPolyOrder, 'MeLOn: SFFT')
            _hdl[0].header['BGORDER'] = (BGPolyOrder, 'MeLOn: SFFT')
            _hdl[0].header['CPHOTR'] = (str(ConstPhotRatio), 'MeLOn: SFFT')
            _hdl[0].header['KERHW'] = (KerHW, 'MeLOn: SFFT')
            _hdl[0].header['CONVD'] = (ConvdSide, 'MeLOn: SFFT')
            _hdl.writeto(FITS_DIFF, overwrite=True)
            _hdl.close()
        
        # * Save solution array
        if FITS_Solution is not None:
            phdu = fits.PrimaryHDU()
            phdu.header['N0'] = (SFFTConfig[0]['N0'], 'MeLOn: SFFT')
            phdu.header['N1'] = (SFFTConfig[0]['N1'], 'MeLOn: SFFT')
            phdu.header['DK'] = (SFFTConfig[0]['DK'], 'MeLOn: SFFT')
            phdu.header['DB'] = (SFFTConfig[0]['DB'], 'MeLOn: SFFT')
            phdu.header['L0'] = (SFFTConfig[0]['L0'], 'MeLOn: SFFT')
            phdu.header['L1'] = (SFFTConfig[0]['L1'], 'MeLOn: SFFT')
            
            phdu.header['FIJ'] = (SFFTConfig[0]['Fij'], 'MeLOn: SFFT')
            phdu.header['FAB'] = (SFFTConfig[0]['Fab'], 'MeLOn: SFFT')
            phdu.header['FPQ'] = (SFFTConfig[0]['Fpq'], 'MeLOn: SFFT')
            phdu.header['FIJAB'] = (SFFTConfig[0]['Fijab'], 'MeLOn: SFFT')

            PixA_Solution = Solution.reshape((-1, 1))
            phdu.data = PixA_Solution.T
            fits.HDUList([phdu]).writeto(FITS_Solution, overwrite=True)
        
        return Solution, PixA_DIFF



import os
import math
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from tempfile import mkdtemp
from sfft.utils.pyAstroMatic.PYSWarp import PY_SWarp
# version: Apr 22, 2024

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Image_ZoomRotate:
    @staticmethod
    def IZR(PixA_obj, ZOOM_SCALE_X=1.0, ZOOM_SCALE_Y=1.0, OUTSIZE_PARIRY_X='UNCHANGED', OUTSIZE_PARIRY_Y='UNCHANGED', \
        PATTERN_ROTATE_ANGLE=0.0, RESAMPLING_TYPE='LANCZOS4', FILL_VALUE=0.0, VERBOSE_LEVEL=2):
        
        """
        # Remarks on Image Zoom & Rotate
        # [1] In this function, 'Zoom' means to change the spatial resolution (pixel size) of given image.
        #     That is to say, we would like to resample the original image on a new grid with different resolution.
        #     The arguments -ZOOM_SCALE_X and -ZOOM_SCALE_Y control the resacle factor of pixel size.
        #     E.g., ZOOM_SCALE_X = 1.2 means new pixel is 1.2 times larger than the original pixel along the X axis.
        #
        # [2] Throughout the image Zoom, the position of image center is always pinned.
        #     The Zoomed image frame has the smallest area which can fully cover the original frame.
        #     NOTE: the image Zoom would not alert the even/odd property of image size.
        # 
        # [3] This function also allows for an additional pattern rotation about the pinned image center.
        #     The argument -PATTERN_ROTATE_ANGLE is the rotate angle (counterclockwise, in deg).
        # 
        # [4] Given that the function is commonly used to manipulate PSF model, 
        #     here we set LANCZOS4 as default resampling method following PSFEx.
        #
        """

        # check inputs
        assert ZOOM_SCALE_X > 0.0
        assert ZOOM_SCALE_Y > 0.0
        assert 0.0 <= PATTERN_ROTATE_ANGLE < 360.0

        # create temporary directory
        TDIR = mkdtemp(suffix=None, prefix=None, dir=None)

        # * create a header with initialized wcs for the original image
        NAXIS1_ORI, NAXIS2_ORI = PixA_obj.shape                           # image original size
        CRPIX1_ORI, CRPIX2_ORI = 0.5+NAXIS1_ORI/2.0, 0.5+NAXIS2_ORI/2.0   # reference point at image center 
        CRVAL1_ORI, CRVAL2_ORI = 180.0, 0.0                               # trivial skycoord of the reference point
        CD1_1_ORI, CD1_2_ORI = 1.0/3600, 0.0                              # trivial pixel scale of the original image
        CD2_1_ORI, CD2_2_ORI = 0.0, 1.0/3600                              # trivial pixel scale of the original image

        wcsinfo_ORI = {'NAXIS': 2,
                       'NAXIS1': NAXIS1_ORI, 'NAXIS2': NAXIS2_ORI,
                       'CRPIX1': CRPIX1_ORI, 'CRPIX2': CRPIX2_ORI,
                       'CRVAL1': CRVAL1_ORI, 'CRVAL2': CRVAL2_ORI,
                       'CD1_1': CD1_1_ORI, 'CD1_2': CD1_2_ORI,
                       'CD2_1': CD2_1_ORI, 'CD2_2': CD2_2_ORI,
                       'CUNIT1': 'deg', 'CUNIT2': 'deg',
                       'CTYPE1': 'RA---TAN', 'CTYPE2': 'DEC--TAN'}

        # * save the original image with initialized wcs
        header_ORI = WCS(wcsinfo_ORI).to_header(relax=True)
        hdl_ORI = fits.HDUList([fits.PrimaryHDU(data=PixA_obj.T, header=header_ORI)])
        
        hdl_ORI[0].header['GAIN'] = (1.0, 'MeLOn: PlaceHolder')    
        hdl_ORI[0].header['SATURATE'] = (50000.0, 'MeLOn: PlaceHolder')
        hdl_ORI[0].header['EXPTIME'] = (60.0, 'MeLOn: PlaceHolder')
        hdl_ORI[0].header['MJD-OBS'] = (58849.0, 'MeLOn: PlaceHolder')
        
        FITS_ORI = TDIR + '/original_image.fits'
        hdl_ORI.writeto(FITS_ORI, overwrite=True)

        # * define the target wcs
        wcsinfo_TARG = wcsinfo_ORI.copy()

        # * determine the number of pixels of the zoomed image
        def NPIX_ORI2ZOOMED(NPIX_ORI, ZOOM_SCALE, OUTSIZE_PARIRY):
            if OUTSIZE_PARIRY == 'UNCHANGED':
                if NPIX_ORI % 2 == 0: OUTSIZE_UPARIRY = 'EVEN'
                if NPIX_ORI % 2 == 1: OUTSIZE_UPARIRY = 'ODD'
            if OUTSIZE_PARIRY == 'ODD': OUTSIZE_UPARIRY = 'ODD'
            if OUTSIZE_PARIRY == 'EVEN': OUTSIZE_UPARIRY = 'EVEN'
            HALFWIDTH_ORI = NPIX_ORI/2.0  # unit of original pixel
            if OUTSIZE_UPARIRY == 'EVEN':
                NPIX_ZOOMED = 2 * math.ceil(HALFWIDTH_ORI/ZOOM_SCALE)
            if OUTSIZE_UPARIRY == 'ODD':
                NPIX_ZOOMED = 2 * math.ceil((HALFWIDTH_ORI - ZOOM_SCALE/2.0)/ZOOM_SCALE) + 1
            return NPIX_ZOOMED

        NAXIS1_ZOOMED = NPIX_ORI2ZOOMED(NAXIS1_ORI, ZOOM_SCALE_X, OUTSIZE_PARIRY_X)
        NAXIS2_ZOOMED = NPIX_ORI2ZOOMED(NAXIS2_ORI, ZOOM_SCALE_Y, OUTSIZE_PARIRY_Y)
        
        # * give the coordinate change due to the frame transform from original to zoomed
        def COORD_ORI2ZOOMED(X_ORI, Y_ORI):
            # NOTE: X_ORI/Y_ORI can be a number or a 1d-array
            OFFSET_X = ZOOM_SCALE_X * (0.5 + NAXIS1_ZOOMED/2.0) - (0.5 + NAXIS1_ORI/2.0)   # unit of original pixel
            OFFSET_Y = ZOOM_SCALE_Y * (0.5 + NAXIS2_ZOOMED/2.0) - (0.5 + NAXIS2_ORI/2.0)   # unit of original pixel
            X_ZOOMED = (X_ORI + OFFSET_X) / ZOOM_SCALE_X
            Y_ZOOMED = (Y_ORI + OFFSET_Y) / ZOOM_SCALE_Y
            return X_ZOOMED, Y_ZOOMED

        # * modify wcs to adapt to the zoomed frame
        CRPIX1_ZOOMED, CRPIX2_ZOOMED = COORD_ORI2ZOOMED(CRPIX1_ORI, CRPIX2_ORI)
        CD1_1_ZOOMED, CD1_2_ZOOMED = ZOOM_SCALE_X * CD1_1_ORI, 0.0
        CD2_1_ZOOMED, CD2_2_ZOOMED = 0.0, ZOOM_SCALE_Y * CD2_2_ORI

        if ZOOM_SCALE_X != 1.0 or ZOOM_SCALE_Y != 1.0:
            if VERBOSE_LEVEL in [1, 2]: 
                print('MeLOn CheckPoint: Modify WCS to adapt to the Zoomed Frame.')
            wcsinfo_TARG['NAXIS1'], wcsinfo_TARG['NAXIS2'] = NAXIS1_ZOOMED, NAXIS2_ZOOMED
            wcsinfo_TARG['CRPIX1'], wcsinfo_TARG['CRPIX2'] = CRPIX1_ZOOMED, CRPIX2_ZOOMED
            wcsinfo_TARG['CD1_1'], wcsinfo_TARG['CD1_2'] = CD1_1_ZOOMED, CD1_2_ZOOMED
            wcsinfo_TARG['CD2_1'], wcsinfo_TARG['CD2_2'] = CD2_1_ZOOMED, CD2_2_ZOOMED
        else:
            if VERBOSE_LEVEL in [1, 2]:
                print('MeLOn CheckPoint: NO IMAGE ZOOM!')
            assert np.allclose([wcsinfo_TARG['NAXIS1'], wcsinfo_TARG['NAXIS2']], [NAXIS1_ZOOMED, NAXIS2_ZOOMED])
            assert np.allclose([wcsinfo_TARG['CRPIX1'], wcsinfo_TARG['CRPIX2']], [CRPIX1_ZOOMED, CRPIX2_ZOOMED])
            assert np.allclose([wcsinfo_TARG['CD1_1'], wcsinfo_TARG['CD1_2']], [CD1_1_ZOOMED, CD1_2_ZOOMED])
            assert np.allclose([wcsinfo_TARG['CD2_1'], wcsinfo_TARG['CD2_2']], [CD2_1_ZOOMED, CD2_2_ZOOMED])

        alpha = np.deg2rad(PATTERN_ROTATE_ANGLE)
        RotMAT = np.array([[np.cos(alpha), -np.sin(alpha)], [np.sin(alpha), np.cos(alpha)]])
        RotMAT_inv = np.array([[np.cos(alpha), np.sin(alpha)], [-np.sin(alpha), np.cos(alpha)]])

        # * give the coordinate change due to pattern rotate on zoomed
        def COORD_ZOOMED2ROTATED(X_ZOOMED, Y_ZOOMED):
            # NOTE: X_ZOOMED/Y_ZOOMED can be a number or a 1d-array
            cX_ZOOMED = X_ZOOMED - CRPIX1_ZOOMED
            cY_ZOOMED = Y_ZOOMED - CRPIX2_ZOOMED
            cCOORD_ZOOMED = np.array([cX_ZOOMED, cY_ZOOMED]).reshape((2, -1))
            cCOORD_ROTATED = np.dot(RotMAT, cCOORD_ZOOMED)
            if cCOORD_ROTATED.shape[1] == 1: cX_ROTATED, cY_ROTATED = cCOORD_ROTATED[:, 0]
            else: cX_ROTATED, cY_ROTATED = cCOORD_ROTATED
            X_ROTATED = cX_ROTATED + CRPIX1_ZOOMED
            Y_ROTATED = cY_ROTATED + CRPIX2_ZOOMED
            return X_ROTATED, Y_ROTATED

        # * further modify wcs to adapt to the rotated pattern
        CDMAT_ZOOMED = np.array([[CD1_1_ZOOMED, CD1_2_ZOOMED], [CD2_1_ZOOMED, CD2_2_ZOOMED]])
        CDMAT_ROTATED = np.dot(CDMAT_ZOOMED, RotMAT_inv)
        CD1_1_ROTATED, CD1_2_ROTATED = CDMAT_ROTATED[0, :]
        CD2_1_ROTATED, CD2_2_ROTATED = CDMAT_ROTATED[1, :]

        if PATTERN_ROTATE_ANGLE != 0.0:
            if VERBOSE_LEVEL in [1, 2]:
                print('MeLOn CheckPoint: Modify WCS to adapt to the Rotated Frame.')
            wcsinfo_TARG['CD1_1'], wcsinfo_TARG['CD1_2'] = CD1_1_ROTATED, CD1_2_ROTATED
            wcsinfo_TARG['CD2_1'], wcsinfo_TARG['CD2_2'] = CD2_1_ROTATED, CD2_2_ROTATED
        else:
            if VERBOSE_LEVEL in [1, 2]:
                print('MeLOn CheckPoint: NO IMAGE ROTATION!')
            assert np.allclose([wcsinfo_TARG['CD1_1'], wcsinfo_TARG['CD1_2']], [CD1_1_ROTATED, CD1_2_ROTATED])
            assert np.allclose([wcsinfo_TARG['CD2_1'], wcsinfo_TARG['CD2_2']], [CD2_1_ROTATED, CD2_2_ROTATED])

        # * give the throughout coordinate change from original to rotated.
        def func_coordtrans(X_ORI, Y_ORI):
            X_ZOOMED, Y_ZOOMED = COORD_ORI2ZOOMED(X_ORI, Y_ORI)
            X_ROTATED, Y_ROTATED = COORD_ZOOMED2ROTATED(X_ZOOMED, Y_ZOOMED)
            return X_ROTATED, Y_ROTATED

        # * save an empty image with target wcs
        header_TARG = WCS(wcsinfo_TARG).to_header(relax=True)
        PixA_empty = np.zeros((wcsinfo_TARG['NAXIS1'], wcsinfo_TARG['NAXIS2'])).astype(float)
        hdl_TARG = fits.HDUList([fits.PrimaryHDU(data=PixA_empty.T, header=header_TARG)])
        
        hdl_TARG[0].header['GAIN'] = (1.0, 'MeLOn: PlaceHolder')    
        hdl_TARG[0].header['SATURATE'] = (50000.0, 'MeLOn: PlaceHolder')   
        hdl_TARG[0].header['EXPTIME'] = (60.0, 'MeLOn: PlaceHolder')
        hdl_TARG[0].header['MJD-OBS'] = (58849.0, 'MeLOn: PlaceHolder')

        FITS_TARG = TDIR + '/target_empty_image.fits'
        hdl_TARG.writeto(FITS_TARG, overwrite=True)

        # * run SWarp to perform the image resampling
        FITS_resamp = TDIR + '/resampled_image.fits'
        PY_SWarp.PS(FITS_obj=FITS_ORI, FITS_ref=FITS_TARG, FITS_resamp=FITS_resamp, \
            GAIN_KEY='GAIN', SATUR_KEY='SATURATE', OVERSAMPLING=1, RESAMPLING_TYPE=RESAMPLING_TYPE, \
            SUBTRACT_BACK='N', FILL_VALUE=FILL_VALUE, VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        PixA_resamp = fits.getdata(FITS_resamp, ext=0).T
        os.system('rm -rf %s' %TDIR)
        
        return PixA_resamp, func_coordtrans



import time
import numpy as np

__last_update__ = "2024-09-21"
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6.0"

class ElementalSFFTSubtract_Cupy:
    @staticmethod
    def ESSC(PixA_I, PixA_J, SFFTConfig, SFFTSolution=None, Subtract=False, VERBOSE_LEVEL=2):
        
        import cupy as cp
        import cupyx.scipy.linalg as cpx_linalg

        def LSSolver(LHMAT_GPU, RHb_GPU):
            """
            # NOTE: cupy/numpy (moderately/extremely) slow version has been DEPRECATED!
            Solution_GPU = cp.linalg.solve(LHMAT_GPU, RHb_GPU)
            Solution_GPU = cp.array(np.linalg.solve(cp.asnumpy(LHMAT_GPU), cp.asnumpy(RHb_GPU)), dtype=np.float64)
            """
            lu_piv_GPU = cpx_linalg.lu_factor(LHMAT_GPU, overwrite_a=False, check_finite=True)
            Solution_GPU = cpx_linalg.lu_solve(lu_piv_GPU, RHb_GPU)
            return Solution_GPU
        
        ta = time.time()
        # * Read SFFT parameters
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --||--||--||--||-- TRIGGER SFFT SUBTRACTION --||--||--||--||-- ')
            print('\n ---||--- KerPolyOrder %d | BGPolyOrder %d | KerHW [%d] ---||--- '\
                   %(SFFTConfig[0]['DK'], SFFTConfig[0]['DB'], SFFTConfig[0]['w0']))

        SFFTParam_dict, SFFTModule_dict = SFFTConfig
        N0, N1 = SFFTParam_dict['N0'], SFFTParam_dict['N1']
        w0, w1 = SFFTParam_dict['w0'], SFFTParam_dict['w1']
        DK, DB = SFFTParam_dict['DK'], SFFTParam_dict['DB']

        if PixA_I.shape != (N0, N1) or PixA_J.shape != (N0, N1):
            _error_message = 'INCONSISTENT shape of input images I & J, [%d, %d] required!' %(N0, N1)
            raise Exception('MeLOn ERROR: %s' %_error_message)

        ConstPhotRatio = SFFTParam_dict['ConstPhotRatio']
        MaxThreadPerB = SFFTParam_dict['MaxThreadPerB']
        
        L0, L1 = SFFTParam_dict['L0'], SFFTParam_dict['L1']
        Fab, Fij, Fpq = SFFTParam_dict['Fab'], SFFTParam_dict['Fij'], SFFTParam_dict['Fpq']
        SCALE, SCALE_L = SFFTParam_dict['SCALE'], SFFTParam_dict['SCALE_L']

        NEQ, Fijab = SFFTParam_dict['NEQ'], SFFTParam_dict['Fijab']
        NEQ_FSfree = SFFTParam_dict['NEQ_FSfree']

        FOMG, FGAM = SFFTParam_dict['FOMG'], SFFTParam_dict['FGAM']
        FTHE, FPSI = SFFTParam_dict['FTHE'], SFFTParam_dict['FPSI']
        FPHI, FDEL = SFFTParam_dict['FPHI'], SFFTParam_dict['FDEL']
        
        # * Grid-Block-Thread Managaement [Pixel-Level]
        GPUManage = lambda NT: ((NT-1)//MaxThreadPerB + 1, min(NT, MaxThreadPerB))
        BpG_PIX0, TpB_PIX0 = GPUManage(N0)
        BpG_PIX1, TpB_PIX1 = GPUManage(N1)
        BpG_PIX, TpB_PIX = (BpG_PIX0, BpG_PIX1), (TpB_PIX0, TpB_PIX1, 1)

        # * Define First-order MultiIndex Reference
        REF_pq = np.array([(p, q) for p in range(DB+1) for q in range(DB+1-p)]).astype(np.int32)
        REF_ij = np.array([(i, j) for i in range(DK+1) for j in range(DK+1-i)]).astype(np.int32)
        REF_ab = np.array([(a_pos-w0, b_pos-w1) for a_pos in range(L0) for b_pos in range(L1)]).astype(np.int32)
        REF_pq_GPU = cp.array(REF_pq)
        REF_ij_GPU = cp.array(REF_ij)
        REF_ab_GPU = cp.array(REF_ab)

        # * Define Second-order MultiIndex Reference
        SREF_iji0j0 = np.array([(ij, i0j0) for ij in range(Fij) for i0j0 in range(Fij)]).astype(np.int32)
        SREF_pqp0q0 = np.array([(pq, p0q0) for pq in range(Fpq) for p0q0 in range(Fpq)]).astype(np.int32)
        SREF_ijpq = np.array([(ij, pq) for ij in range(Fij) for pq in range(Fpq)]).astype(np.int32)
        SREF_pqij = np.array([(pq, ij) for pq in range(Fpq) for ij in range(Fij)]).astype(np.int32)
        SREF_ijab = np.array([(ij, ab) for ij in range(Fij) for ab in range(Fab)]).astype(np.int32)

        SREF_iji0j0_GPU = cp.array(SREF_iji0j0)
        SREF_pqp0q0_GPU = cp.array(SREF_pqp0q0)
        SREF_ijpq_GPU = cp.array(SREF_ijpq)
        SREF_pqij_GPU = cp.array(SREF_pqij)
        SREF_ijab_GPU = cp.array(SREF_ijab)

        # * Define the Forbidden ij-ab Rule
        ij00 = np.arange(w0 * L1 + w1, Fijab, Fab).astype(np.int32)
        if ConstPhotRatio:
            FBijab = ij00[1:]
            MASK_nFS = np.ones(NEQ).astype(bool)
            MASK_nFS[FBijab] = False            
            IDX_nFS = np.where(MASK_nFS)[0].astype(np.int32)
            IDX_nFS_GPU = cp.array(IDX_nFS)
            NEQ_FSfree = len(IDX_nFS)

        t0 = time.time()
        # * Read input images as C-order arrays
        if not PixA_I.flags['C_CONTIGUOUS']:
            PixA_I = np.ascontiguousarray(PixA_I, np.float64)
            PixA_I_GPU = cp.array(PixA_I)
        else: PixA_I_GPU = cp.array(PixA_I.astype(np.float64))
        
        if not PixA_J.flags['C_CONTIGUOUS']:
            PixA_J = np.ascontiguousarray(PixA_J, np.float64)
            PixA_J_GPU = cp.array(PixA_J)
        else: PixA_J_GPU = cp.array(PixA_J.astype(np.float64))
        dt0 = time.time() - t0

        # * Symbol Convention NOTE
        #    X (x) / Y (y) ----- pixel row / column index
        #    CX (cx) / CY (cy) ----- ScaledFortranCoor of pixel (x, y) center   
        #    e.g. pixel (x, y) = (3, 5) corresponds (cx, cy) = (4.0/N0, 6.0/N1)
        #    NOTE cx / cy is literally \mathtt{x} / \mathtt{y} in sfft paper.
        #    NOTE Without special definition, MeLOn convention refers to X (x) / Y (y) as FortranCoor.

        # * Get Spatial Coordinates
        t1 = time.time()
        PixA_X_GPU = cp.zeros((N0, N1), dtype=np.int32)      # row index, [0, N0)
        PixA_Y_GPU = cp.zeros((N0, N1), dtype=np.int32)      # column index, [0, N1)
        PixA_CX_GPU = cp.zeros((N0, N1), dtype=np.float64)   # coordinate.x
        PixA_CY_GPU = cp.zeros((N0, N1), dtype=np.float64)   # coordinate.y 

        _module = SFFTModule_dict['SpatialCoor']
        _func = _module.get_function('kmain')
        _func(args=(PixA_X_GPU, PixA_Y_GPU, PixA_CX_GPU, PixA_CY_GPU), block=TpB_PIX, grid=BpG_PIX)

        # * Spatial Polynomial terms Iij, Tpq
        SPixA_Iij_GPU = cp.zeros((Fij, N0, N1), dtype=np.float64)
        SPixA_Tpq_GPU = cp.zeros((Fpq, N0, N1), dtype=np.float64)

        _module = SFFTModule_dict['SpatialPoly']
        _func = _module.get_function('kmain')
        _func(args=(REF_ij_GPU, REF_pq_GPU, PixA_CX_GPU, PixA_CY_GPU, PixA_I_GPU, \
            SPixA_Iij_GPU, SPixA_Tpq_GPU), block=TpB_PIX, grid=BpG_PIX)
        
        del PixA_I_GPU
        dt1 = time.time() - t1

        t2 = time.time()

        """
        # alternative way to perform FFT
        import cupyx as cpx
        plan = cpx.scipy.fftpack.get_fft_plan(x)
        y = cpx.scipy.fftpack.fftn(x, plan=plan)
        
        """

        # * Make DFT of J, Iij, Tpq and their conjugates
        PixA_FJ_GPU = cp.empty((N0, N1), dtype=np.complex128)
        PixA_FJ_GPU[:, :] = PixA_J_GPU.astype(np.complex128)
        PixA_FJ_GPU[:, :] = cp.fft.fft2(PixA_FJ_GPU)
        PixA_FJ_GPU[:, :] *= SCALE

        SPixA_FIij_GPU = cp.empty((Fij, N0, N1), dtype=np.complex128)
        SPixA_FIij_GPU[:, :, :] = SPixA_Iij_GPU.astype(np.complex128)
        for k in range(Fij): 
            SPixA_FIij_GPU[k: k+1] = cp.fft.fft2(SPixA_FIij_GPU[k: k+1])
        SPixA_FIij_GPU[:, :] *= SCALE

        SPixA_FTpq_GPU = cp.empty((Fpq, N0, N1), dtype=np.complex128)
        SPixA_FTpq_GPU[:, :, :] = SPixA_Tpq_GPU.astype(np.complex128)
        for k in range(Fpq): 
            SPixA_FTpq_GPU[k: k+1] = cp.fft.fft2(SPixA_FTpq_GPU[k: k+1])
        SPixA_FTpq_GPU[:, :] *= SCALE

        del SPixA_Iij_GPU
        del SPixA_Tpq_GPU

        PixA_CFJ_GPU = cp.conj(PixA_FJ_GPU)
        SPixA_CFIij_GPU = cp.conj(SPixA_FIij_GPU)
        SPixA_CFTpq_GPU = cp.conj(SPixA_FTpq_GPU)
        dt2 = time.time() - t2
        dta = time.time() - ta

        if VERBOSE_LEVEL in [1, 2]:
            print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Preliminary Steps takes [%.4fs]' %dta)

        if VERBOSE_LEVEL in [2]:
            print('/////   a   ///// Read Input Images  (%.4fs)' %dt0)
            print('/////   b   ///// Spatial Polynomial (%.4fs)' %dt1)
            print('/////   c   ///// DFT-%d             (%.4fs)' %(1 + Fij + Fpq, dt2))

        # * Consider The Major Sources of the Linear System 
        #     𝛀_i8j8ij     &    𝜦_i8j8pq    ||     𝚯_i8j8
        #     𝜳_p8q8ij     &    𝚽_p8q8pq    ||     𝚫_p8q8   
        #
        # * Remarks
        #    a. They have consistent form: Greek(rho, eps) = PreGreek(Mod_N0(rho), Mod_N1(eps))
        #    b. PreGreek = s * Re[DFT(HpGreek)], where HpGreek = GLH * GRH and s is some real-scale.
        #    c. Considering the subscripted variables, HpGreek / PreGreek is Complex / Real 3D with shape (F_Greek, N0, N1).
        
        if SFFTSolution is not None:
            Solution = SFFTSolution 
            Solution_GPU = cp.array(Solution.astype(np.float64))
            a_ijab_GPU = Solution_GPU[: Fijab]
            b_pq_GPU = Solution_GPU[Fijab: ]

        if SFFTSolution is None:
            tb = time.time()
            # * Grid-Block-Thread Managaement [for Greek]
            BpG_OMG0, TpB_OMG0 = GPUManage(Fijab)
            BpG_OMG1, TpB_OMG1 = GPUManage(Fijab)
            BpG_OMG, TpB_OMG = (BpG_OMG0, BpG_OMG1), (TpB_OMG0, TpB_OMG1, 1)

            BpG_GAM0, TpB_GAM0 = GPUManage(Fijab)
            BpG_GAM1, TpB_GAM1 = GPUManage(Fpq)
            BpG_GAM, TpB_GAM = (BpG_GAM0, BpG_GAM1), (TpB_GAM0, TpB_GAM1, 1)

            BpG_THE0, TpB_THE0 = GPUManage(Fijab)
            BpG_THE, TpB_THE = (BpG_THE0, 1), (TpB_THE0, 1, 1)

            BpG_PSI0, TpB_PSI0 = GPUManage(Fpq)
            BpG_PSI1, TpB_PSI1 = GPUManage(Fijab)
            BpG_PSI, TpB_PSI = (BpG_PSI0, BpG_PSI1), (TpB_PSI0, TpB_PSI1, 1)

            BpG_PHI0, TpB_PHI0 = GPUManage(Fpq)
            BpG_PHI1, TpB_PHI1 = GPUManage(Fpq)
            BpG_PHI, TpB_PHI = (BpG_PHI0, BpG_PHI1), (TpB_PHI0, TpB_PHI1, 1)

            BpG_DEL0, TpB_DEL0 = GPUManage(Fpq)
            BpG_DEL, TpB_DEL = (BpG_DEL0, 1), (TpB_DEL0, 1, 1)

            LHMAT_GPU = cp.empty((NEQ, NEQ), dtype=np.float64)
            RHb_GPU = cp.empty(NEQ, dtype=np.float64)
            t3 = time.time()

            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝛀  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝛀 [HpOMG]
            _module = SFFTModule_dict['HadProd_OMG']
            _func = _module.get_function('kmain')
            HpOMG_GPU = cp.empty((FOMG, N0, N1), dtype=np.complex128)
            _func(args=(SREF_iji0j0_GPU, SPixA_FIij_GPU, SPixA_CFIij_GPU, HpOMG_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PreOMG = SCALE * Re[DFT(HpOMG)]
            for k in range(FOMG):
                HpOMG_GPU[k: k+1] = cp.fft.fft2(HpOMG_GPU[k: k+1])
            HpOMG_GPU *= SCALE

            PreOMG_GPU = cp.empty((FOMG, N0, N1), dtype=np.float64)
            PreOMG_GPU[:, :, :] = HpOMG_GPU.real
            PreOMG_GPU[:, :, :] *= SCALE
            del HpOMG_GPU

            # c. Fill Linear System with PreOMG
            _module = SFFTModule_dict['FillLS_OMG']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PreOMG_GPU, LHMAT_GPU), \
                block=TpB_OMG, grid=BpG_OMG)

            del PreOMG_GPU
            dt3 = time.time() - t3

            t4 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝜦  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝜦 [HpGAM]
            _module = SFFTModule_dict['HadProd_GAM']
            _func = _module.get_function('kmain')
            HpGAM_GPU = cp.empty((FGAM, N0, N1), dtype=np.complex128)
            _func(args=(SREF_ijpq_GPU, SPixA_FIij_GPU, SPixA_CFTpq_GPU, HpGAM_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PreGAM = 1 * Re[DFT(HpGAM)]
            for k in range(FGAM):
                HpGAM_GPU[k: k+1] = cp.fft.fft2(HpGAM_GPU[k: k+1])
            HpGAM_GPU *= SCALE

            PreGAM_GPU = cp.empty((FGAM, N0, N1), dtype=np.float64)
            PreGAM_GPU[:, :, :] = HpGAM_GPU.real
            del HpGAM_GPU

            # c. Fill Linear System with PreGAM
            _module = SFFTModule_dict['FillLS_GAM']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PreGAM_GPU, LHMAT_GPU), \
                block=TpB_GAM, grid=BpG_GAM)

            del PreGAM_GPU
            dt4 = time.time() - t4

            t5 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝜳  -- -- -- -- -- -- -- -- *
            
            # a. Hadamard Product for 𝜳  [HpPSI]
            _module = SFFTModule_dict['HadProd_PSI']
            _func = _module.get_function('kmain')
            HpPSI_GPU = cp.empty((FPSI, N0, N1), dtype=np.complex128)
            _func(args=(SREF_pqij_GPU, SPixA_CFIij_GPU, SPixA_FTpq_GPU, HpPSI_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PrePSI = 1 * Re[DFT(HpPSI)]
            for k in range(FPSI):
                HpPSI_GPU[k: k+1] = cp.fft.fft2(HpPSI_GPU[k: k+1])
            HpPSI_GPU *= SCALE

            PrePSI_GPU = cp.empty((FPSI, N0, N1), dtype=np.float64)
            PrePSI_GPU[:, :, :] = HpPSI_GPU.real
            del HpPSI_GPU

            # c. Fill Linear System with PrePSI
            _module = SFFTModule_dict['FillLS_PSI']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PrePSI_GPU, LHMAT_GPU), \
                block=TpB_PSI, grid=BpG_PSI)

            del PrePSI_GPU
            dt5 = time.time() - t5

            t6 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝚽  -- -- -- -- -- -- -- -- *
            
            # a. Hadamard Product for 𝚽  [HpPHI]
            _module = SFFTModule_dict['HadProd_PHI']
            _func = _module.get_function('kmain')
            HpPHI_GPU = cp.empty((FPHI, N0, N1), dtype=np.complex128)
            _func(args=(SREF_pqp0q0_GPU, SPixA_FTpq_GPU, SPixA_CFTpq_GPU, HpPHI_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PrePHI = SCALE_L * Re[DFT(HpPHI)]
            for k in range(FPHI):
                HpPHI_GPU[k: k+1] = cp.fft.fft2(HpPHI_GPU[k: k+1])
            HpPHI_GPU *= SCALE

            PrePHI_GPU = cp.empty((FPHI, N0, N1), dtype=np.float64)
            PrePHI_GPU[:, :, :] = HpPHI_GPU.real
            PrePHI_GPU[:, :, :] *= SCALE_L
            del HpPHI_GPU

            # c. Fill Linear System with PrePHI
            _module = SFFTModule_dict['FillLS_PHI']
            _func = _module.get_function('kmain')
            _func(args=(PrePHI_GPU, LHMAT_GPU), block=TpB_PHI, grid=BpG_PHI)

            del PrePHI_GPU
            dt6 = time.time() - t6

            t7 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝚯 & 𝚫  -- -- -- -- -- -- -- -- *

            # a1. Hadamard Product for 𝚯  [HpTHE]
            _module = SFFTModule_dict['HadProd_THE']
            _func = _module.get_function('kmain')
            HpTHE_GPU = cp.empty((FTHE, N0, N1), dtype=np.complex128)
            _func(args=(SPixA_FIij_GPU, PixA_CFJ_GPU, HpTHE_GPU), block=TpB_PIX, grid=BpG_PIX)

            # a2. Hadamard Product for 𝚫  [HpDEL]
            _module = SFFTModule_dict['HadProd_DEL']
            _func = _module.get_function('kmain')
            HpDEL_GPU = cp.empty((FDEL, N0, N1), dtype=np.complex128)
            _func(args=(SPixA_FTpq_GPU, PixA_CFJ_GPU, HpDEL_GPU), block=TpB_PIX, grid=BpG_PIX)

            # b1. PreTHE = 1 * Re[DFT(HpTHE)]
            # b2. PreDEL = SCALE_L * Re[DFT(HpDEL)]
            for k in range(FTHE):
                HpTHE_GPU[k: k+1] = cp.fft.fft2(HpTHE_GPU[k: k+1])
            HpTHE_GPU[:, :, :] *= SCALE
            
            for k in range(FDEL):
                HpDEL_GPU[k: k+1] = cp.fft.fft2(HpDEL_GPU[k: k+1])
            HpDEL_GPU[:, :, :] *= SCALE

            PreTHE_GPU = cp.empty((FTHE, N0, N1), dtype=np.float64)
            PreTHE_GPU[:, :, :] = HpTHE_GPU.real
            del HpTHE_GPU
            
            PreDEL_GPU = cp.empty((FDEL, N0, N1), dtype=np.float64)
            PreDEL_GPU[:, :, :] = HpDEL_GPU.real
            PreDEL_GPU[:, :, :] *= SCALE_L
            del HpDEL_GPU
            
            # c1. Fill Linear System with PreTHE
            _module = SFFTModule_dict['FillLS_THE']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PreTHE_GPU, RHb_GPU), \
                block=TpB_THE, grid=BpG_THE)
            del PreTHE_GPU

            # c2. Fill Linear System with PreDEL
            _module = SFFTModule_dict['FillLS_DEL']
            _func = _module.get_function('kmain')
            _func(args=(PreDEL_GPU, RHb_GPU), block=TpB_DEL, grid=BpG_DEL)
            
            del PreDEL_GPU
            dt7 = time.time() - t7

            # * -- -- -- -- -- -- -- -- Remove Forbidden Stripes  -- -- -- -- -- -- -- -- *
            if not ConstPhotRatio: pass
            if ConstPhotRatio:
                RHb_FSfree_GPU = cp.take(RHb_GPU, IDX_nFS_GPU)
                _module = SFFTModule_dict['Remove_LSFStripes']
                _func = _module.get_function('kmain')
                BpG_FSfree_PA, TpB_FSfree_PA = GPUManage(NEQ_FSfree)     # Per Axis
                BpG_FSfree, TpB_FSfree = (BpG_FSfree_PA, BpG_FSfree_PA), (TpB_FSfree_PA, TpB_FSfree_PA, 1)
                LHMAT_FSfree_GPU = cp.empty((NEQ_FSfree, NEQ_FSfree), dtype=np.float64)
                _func(args=(LHMAT_GPU, IDX_nFS_GPU, LHMAT_FSfree_GPU), block=TpB_FSfree, grid=BpG_FSfree)

            t8 = time.time()
            # * -- -- -- -- -- -- -- -- Solve Linear System  -- -- -- -- -- -- -- -- *
            if not ConstPhotRatio: 
                Solution_GPU = LSSolver(LHMAT_GPU=LHMAT_GPU, RHb_GPU=RHb_GPU)

            if ConstPhotRatio:
                # Extend the solution to be consistent form
                Solution_FSfree_GPU = LSSolver(LHMAT_GPU=LHMAT_FSfree_GPU, RHb_GPU=RHb_FSfree_GPU)
                _module = SFFTModule_dict['Extend_Solution']
                _func = _module.get_function('kmain')
                BpG_ES, TpB_ES = (BpG_FSfree_PA, 1), (TpB_FSfree_PA, 1, 1)
                Solution_GPU = cp.zeros(NEQ, dtype=np.float64)
                _func(args=(Solution_FSfree_GPU, IDX_nFS_GPU, Solution_GPU), block=TpB_ES, grid=BpG_ES)
            
            a_ijab_GPU = Solution_GPU[: Fijab]
            b_pq_GPU = Solution_GPU[Fijab: ]
            Solution = cp.asnumpy(Solution_GPU)
            dt8 = time.time() - t8
            dtb = time.time() - tb

            if VERBOSE_LEVEL in [1, 2]:
                print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Establish & Solve Linear System takes [%.4fs]' %dtb)

            if VERBOSE_LEVEL in [2]:
                print('/////   d   ///// Establish OMG                       (%.4fs)' %dt3)
                print('/////   e   ///// Establish GAM                       (%.4fs)' %dt4)
                print('/////   f   ///// Establish PSI                       (%.4fs)' %dt5)
                print('/////   g   ///// Establish PHI                       (%.4fs)' %dt6)
                print('/////   h   ///// Establish THE & DEL                 (%.4fs)' %dt7)
                print('/////   i   ///// Solve Linear System                 (%.4fs)' %dt8)

        # * Perform Subtraction 
        PixA_DIFF = None
        if Subtract:
            tc = time.time()
            t9 = time.time()
            # Calculate Kab components
            Wl_GPU = cp.exp((-2j*np.pi/N0) * PixA_X_GPU.astype(np.float64))    # row index l, [0, N0)
            Wm_GPU = cp.exp((-2j*np.pi/N1) * PixA_Y_GPU.astype(np.float64))    # column index m, [0, N1)
            Kab_Wla_GPU = cp.empty((L0, N0, N1), dtype=np.complex128)
            Kab_Wmb_GPU = cp.empty((L1, N0, N1), dtype=np.complex128)

            if w0 == w1:
                wx = w0   # a little bit faster
                for aob in range(-wx, wx+1):
                    Kab_Wla_GPU[aob + wx] = Wl_GPU ** aob    # offset 
                    Kab_Wmb_GPU[aob + wx] = Wm_GPU ** aob    # offset 
            else:
                for a in range(-w0, w0+1): 
                    Kab_Wla_GPU[a + w0] = Wl_GPU ** a        # offset 
                for b in range(-w1, w1+1): 
                    Kab_Wmb_GPU[b + w1] = Wm_GPU ** b        # offset 
            dt9 = time.time() - t9

            t10 = time.time()
            # Construct Difference in Fourier Space
            _module = SFFTModule_dict['Construct_FDIFF']
            _func = _module.get_function('kmain')     
            PixA_FDIFF_GPU = cp.empty((N0, N1), dtype=np.complex128)
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, a_ijab_GPU.astype(np.complex128), \
                SPixA_FIij_GPU, Kab_Wla_GPU, Kab_Wmb_GPU, b_pq_GPU.astype(np.complex128), \
                SPixA_FTpq_GPU, PixA_FJ_GPU, PixA_FDIFF_GPU), block=TpB_PIX, grid=BpG_PIX)
            
            # Get Difference & Reconstructed Images
            PixA_DIFF_GPU = SCALE_L * cp.fft.ifft2(PixA_FDIFF_GPU)
            PixA_DIFF = cp.asnumpy(PixA_DIFF_GPU.real)
            dt10 = time.time() - t10
            dtc = time.time() - tc

            if VERBOSE_LEVEL in [1, 2]:
                print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Perform Subtraction takes [%.4fs]' %dtc)
            
            if VERBOSE_LEVEL in [2]:
                print('/////   j   ///// Calculate Kab         (%.4fs)' %dt9)
                print('/////   k   ///// Construct DIFF        (%.4fs)' %dt10)
        
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --||--||--||--||-- EXIT SFFT SUBTRACTION --||--||--||--||-- ')

        return Solution, PixA_DIFF

class ElementalSFFTSubtract_Numpy:
    @staticmethod
    def ESSN(PixA_I, PixA_J, SFFTConfig, SFFTSolution=None, Subtract=False, NUM_CPU_THREADS_4SUBTRACT=8, VERBOSE_LEVEL=2):

        import pyfftw
        pyfftw.config.NUM_THREADS = NUM_CPU_THREADS_4SUBTRACT
        pyfftw.interfaces.cache.enable()
        
        ta = time.time()
        # * Read SFFT parameters
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --||--||--||--||-- TRIGGER SFFT SUBTRACTION --||--||--||--||-- ')
            print('\n ---||--- KerPolyOrder %d | BGPolyOrder %d | KerHW [%d] ---||--- '\
                   %(SFFTConfig[0]['DK'], SFFTConfig[0]['DB'], SFFTConfig[0]['w0']))

        SFFTParam_dict, SFFTModule_dict = SFFTConfig
        N0, N1 = SFFTParam_dict['N0'], SFFTParam_dict['N1']
        w0, w1 = SFFTParam_dict['w0'], SFFTParam_dict['w1']
        DK, DB = SFFTParam_dict['DK'], SFFTParam_dict['DB']

        if PixA_I.shape != (N0, N1) or PixA_J.shape != (N0, N1):
            _error_message = 'INCONSISTENT shape of input images I & J, [%d, %d] required!' %(N0, N1)
            raise Exception('MeLOn ERROR: %s' %_error_message)

        ConstPhotRatio = SFFTParam_dict['ConstPhotRatio']
        L0, L1 = SFFTParam_dict['L0'], SFFTParam_dict['L1']
        Fab, Fij, Fpq = SFFTParam_dict['Fab'], SFFTParam_dict['Fij'], SFFTParam_dict['Fpq']
        SCALE, SCALE_L = SFFTParam_dict['SCALE'], SFFTParam_dict['SCALE_L']

        NEQ, Fijab = SFFTParam_dict['NEQ'], SFFTParam_dict['Fijab']
        NEQ_FSfree = SFFTParam_dict['NEQ_FSfree']

        FOMG, FGAM = SFFTParam_dict['FOMG'], SFFTParam_dict['FGAM']
        FTHE, FPSI = SFFTParam_dict['FTHE'], SFFTParam_dict['FPSI']
        FPHI, FDEL = SFFTParam_dict['FPHI'], SFFTParam_dict['FDEL']

        # * Define First-order MultiIndex Reference
        REF_pq = np.array([(p, q) for p in range(DB+1) for q in range(DB+1-p)]).astype(np.int32)
        REF_ij = np.array([(i, j) for i in range(DK+1) for j in range(DK+1-i)]).astype(np.int32)
        REF_ab = np.array([(a_pos-w0, b_pos-w1) for a_pos in range(L0) for b_pos in range(L1)]).astype(np.int32)

        # * Define Second-order MultiIndex Reference
        SREF_iji0j0 = np.array([(ij, i0j0) for ij in range(Fij) for i0j0 in range(Fij)]).astype(np.int32)
        SREF_pqp0q0 = np.array([(pq, p0q0) for pq in range(Fpq) for p0q0 in range(Fpq)]).astype(np.int32)
        SREF_ijpq = np.array([(ij, pq) for ij in range(Fij) for pq in range(Fpq)]).astype(np.int32)
        SREF_pqij = np.array([(pq, ij) for pq in range(Fpq) for ij in range(Fij)]).astype(np.int32)
        SREF_ijab = np.array([(ij, ab) for ij in range(Fij) for ab in range(Fab)]).astype(np.int32)

        # * Define the Forbidden ij-ab Rule
        ij00 = np.arange(w0 * L1 + w1, Fijab, Fab).astype(np.int32)
        if ConstPhotRatio:
            FBijab = ij00[1:]
            MASK_nFS = np.ones(NEQ).astype(bool)
            MASK_nFS[FBijab] = False            
            IDX_nFS = np.where(MASK_nFS)[0].astype(np.int32)
            NEQ_FSfree = len(IDX_nFS)

        t0 = time.time()
        # * Read input images as C-order arrays
        if not PixA_I.flags['C_CONTIGUOUS']:
            PixA_I = np.ascontiguousarray(PixA_I, np.float64)
        else: PixA_I = PixA_I.astype(np.float64)

        if not PixA_J.flags['C_CONTIGUOUS']:
            PixA_J = np.ascontiguousarray(PixA_J, np.float64)
        else: PixA_J = PixA_J.astype(np.float64)
        dt0 = time.time() - t0

        # * Symbol Convention NOTE
        #    X (x) / Y (y) ----- pixel row / column index
        #    CX (cx) / CY (cy) ----- ScaledFortranCoor of pixel (x, y) center   
        #    e.g. pixel (x, y) = (3, 5) corresponds (cx, cy) = (4.0/N0, 6.0/N1)
        #    NOTE cx / cy is literally \mathtt{x} / \mathtt{y} in sfft paper.
        #    NOTE Without special definition, MeLOn convention refers to X (x) / Y (y) as FortranCoor.

        # * Get Spatial Coordinates
        t1 = time.time()
        PixA_X = np.zeros((N0, N1), dtype=np.int32)      # row index, [0, N0)
        PixA_Y = np.zeros((N0, N1), dtype=np.int32)      # column index, [0, N1)
        PixA_CX = np.zeros((N0, N1), dtype=np.float64)   # coordinate.x
        PixA_CY = np.zeros((N0, N1), dtype=np.float64)   # coordinate.y 

        _func = SFFTModule_dict['SpatialCoor']
        _func(PixA_X=PixA_X, PixA_Y=PixA_Y, PixA_CX=PixA_CX, PixA_CY=PixA_CY)

        # * Spatial Polynomial terms Iij, Tpq
        SPixA_Iij = np.zeros((Fij, N0, N1), dtype=np.float64)
        SPixA_Tpq = np.zeros((Fpq, N0, N1), dtype=np.float64)

        _func = SFFTModule_dict['SpatialPoly']
        _func(REF_ij=REF_ij, REF_pq=REF_pq, PixA_CX=PixA_CX, PixA_CY=PixA_CY, \
            PixA_I=PixA_I, SPixA_Iij=SPixA_Iij, SPixA_Tpq=SPixA_Tpq)
        dt1 = time.time() - t1

        t2 = time.time()
        Batchsize = 1 + Fij + Fpq
        _SPixA_FFT = np.empty((Batchsize, N0, N1), dtype=np.complex128)
        _SPixA_FFT[0, :, :] = PixA_J.astype(np.complex128)
        _SPixA_FFT[1: Fij+1, :, :] = SPixA_Iij.astype(np.complex128)
        _SPixA_FFT[Fij+1:, :, :] = SPixA_Tpq.astype(np.complex128)

        #_SPixA_FFT = SCALE * np.fft.fft2(_SPixA_FFT)
        _SPixA_FFT = SCALE * pyfftw.interfaces.numpy_fft.fft2(_SPixA_FFT)

        PixA_FJ = _SPixA_FFT[0, :, :]
        SPixA_FIij = _SPixA_FFT[1: Fij+1, :, :]
        SPixA_FTpq = _SPixA_FFT[Fij+1:, :, :]

        PixA_CFJ = np.conj(PixA_FJ)
        SPixA_CFIij = np.conj(SPixA_FIij)
        SPixA_CFTpq = np.conj(SPixA_FTpq)
        dt2 = time.time() - t2
        dta = time.time() - ta

        if VERBOSE_LEVEL in [1, 2]:
            print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Preliminary Steps takes [%.4fs]' %dta)

        if VERBOSE_LEVEL in [2]:
            print('/////   a   ///// Read Input Images  (%.4fs)' %dt0)
            print('/////   b   ///// Spatial Polynomial (%.4fs)' %dt1)
            print('/////   c   ///// DFT-%d             (%.4fs)' %(1 + Fij + Fpq, dt2))

        # * Consider The Major Sources of the Linear System 
        #     𝛀_i8j8ij     &    𝜦_i8j8pq    ||     𝚯_i8j8
        #     𝜳_p8q8ij     &    𝚽_p8q8pq    ||     𝚫_p8q8   
        #
        # * Remarks
        #    a. They have consistent form: Greek(rho, eps) = PreGreek(Mod_N0(rho), Mod_N1(eps))
        #    b. PreGreek = s * Re[DFT(HpGreek)], where HpGreek = GLH * GRH and s is some real-scale.
        #    c. Considering the subscripted variables, HpGreek / PreGreek is Complex / Real 3D with shape (F_Greek, N0, N1).

        if SFFTSolution is not None:
            Solution = SFFTSolution.astype(np.float64)
            a_ijab = Solution[: Fijab]
            b_pq = Solution[Fijab: ]
            
        if SFFTSolution is None:
            tb = time.time()    
            
            LHMAT = np.empty((NEQ, NEQ), dtype=np.float64)
            RHb = np.empty(NEQ, dtype=np.float64)
            t3 = time.time()
            
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝛀  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝛀 [HpOMG]
            _func = SFFTModule_dict['HadProd_OMG']
            HpOMG = np.empty((FOMG, N0, N1), dtype=np.complex128)
            HpOMG = _func(SREF_iji0j0=SREF_iji0j0, SPixA_FIij=SPixA_FIij, SPixA_CFIij=SPixA_CFIij, HpOMG=HpOMG)
            
            # b. PreOMG = SCALE * Re[DFT(HpOMG)]
            HpOMG = SCALE * pyfftw.interfaces.numpy_fft.fft2(HpOMG)
            
            HpOMG = HpOMG.real
            HpOMG *= SCALE
            PreOMG = HpOMG
            
            # c. Fill Linear System with PreOMG
            _func = SFFTModule_dict['FillLS_OMG']
            LHMAT = _func(SREF_ijab=SREF_ijab, REF_ab=REF_ab, PreOMG=PreOMG, LHMAT=LHMAT)
            dt3 = time.time() - t3

            t4 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝜦  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝜦 [HpGAM]
            _func = SFFTModule_dict['HadProd_GAM']
            HpGAM = np.empty((FGAM, N0, N1), dtype=np.complex128)
            HpGAM = _func(SREF_ijpq=SREF_ijpq, SPixA_FIij=SPixA_FIij, SPixA_CFTpq=SPixA_CFTpq, HpGAM=HpGAM)

            # b. PreGAM = 1 * Re[DFT(HpGAM)]
            HpGAM = SCALE * pyfftw.interfaces.numpy_fft.fft2(HpGAM)
            
            HpGAM = HpGAM.real 
            PreGAM = HpGAM

            # c. Fill Linear System with PreGAM
            _func = SFFTModule_dict['FillLS_GAM']
            LHMAT = _func(SREF_ijab=SREF_ijab, REF_ab=REF_ab, PreGAM=PreGAM, LHMAT=LHMAT)
            dt4 = time.time() - t4

            t5 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝜳  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝜳  [HpPSI]
            _func = SFFTModule_dict['HadProd_PSI']
            HpPSI = np.empty((FPSI, N0, N1), dtype=np.complex128)
            HpPSI = _func(SREF_pqij=SREF_pqij, SPixA_CFIij=SPixA_CFIij, SPixA_FTpq=SPixA_FTpq, HpPSI=HpPSI)
            
            # b. PrePSI = 1 * Re[DFT(HpPSI)]
            HpPSI = SCALE * pyfftw.interfaces.numpy_fft.fft2(HpPSI)

            HpPSI = HpPSI.real
            PrePSI = HpPSI

            # c. Fill Linear System with PrePSI
            _func = SFFTModule_dict['FillLS_PSI']
            LHMAT = _func(SREF_ijab=SREF_ijab, REF_ab=REF_ab, PrePSI=PrePSI, LHMAT=LHMAT)
            dt5 = time.time() - t5

            t6 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝚽  -- -- -- -- -- -- -- -- *
            
            # a. Hadamard Product for 𝚽  [HpPHI]
            _func = SFFTModule_dict['HadProd_PHI']
            HpPHI = np.empty((FPHI, N0, N1), dtype=np.complex128)
            HpPHI = _func(SREF_pqp0q0=SREF_pqp0q0, SPixA_FTpq=SPixA_FTpq, SPixA_CFTpq=SPixA_CFTpq, HpPHI=HpPHI)

            # b. PrePHI = SCALE_L * Re[DFT(HpPHI)]
            HpPHI = SCALE * pyfftw.interfaces.numpy_fft.fft2(HpPHI)
            
            HpPHI = HpPHI.real
            HpPHI *= SCALE_L
            PrePHI = HpPHI

            # c. Fill Linear System with PrePHI
            _func = SFFTModule_dict['FillLS_PHI']
            LHMAT = _func(PrePHI=PrePHI, LHMAT=LHMAT)

            dt6 = time.time() - t6
            
            t7 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝚯 & 𝚫  -- -- -- -- -- -- -- -- *

            # a1. Hadamard Product for 𝚯  [HpTHE]
            _func = SFFTModule_dict['HadProd_THE']
            HpTHE = np.empty((FTHE, N0, N1), dtype=np.complex128)
            HpTHE = _func(SPixA_FIij=SPixA_FIij, PixA_CFJ=PixA_CFJ, HpTHE=HpTHE)

            # a2. Hadamard Product for 𝚫  [HpDEL]
            _func = SFFTModule_dict['HadProd_DEL']
            HpDEL = np.empty((FDEL, N0, N1), dtype=np.complex128)
            HpDEL = _func(SPixA_FTpq=SPixA_FTpq, PixA_CFJ=PixA_CFJ, HpDEL=HpDEL)
            
            # b1. PreTHE = 1 * Re[DFT(HpTHE)]
            # b2. PreDEL = SCALE_L * Re[DFT(HpDEL)]
            HpTHE = SCALE * pyfftw.interfaces.numpy_fft.fft2(HpTHE)
            HpDEL = SCALE * pyfftw.interfaces.numpy_fft.fft2(HpDEL)
            
            HpTHE = HpTHE.real
            PreTHE = HpTHE

            HpDEL = HpDEL.real
            HpDEL *= SCALE_L
            PreDEL = HpDEL
            
            # c1. Fill Linear System with PreTHE
            _func = SFFTModule_dict['FillLS_THE']
            RHb = _func(SREF_ijab=SREF_ijab, REF_ab=REF_ab, PreTHE=PreTHE, RHb=RHb)

            # c2. Fill Linear System with PreDEL
            _func = SFFTModule_dict['FillLS_DEL']
            RHb = _func(PreDEL=PreDEL, RHb=RHb)
            dt7 = time.time() - t7
            
            t8 = time.time()
            # * -- -- -- -- -- -- -- -- Remove Forbidden Stripes  -- -- -- -- -- -- -- -- *
            if not ConstPhotRatio: pass
            if ConstPhotRatio:
                RHb_FSfree = RHb[IDX_nFS]
                _func = SFFTModule_dict['Remove_LSFStripes']
                LHMAT_FSfree = np.empty((NEQ_FSfree, NEQ_FSfree), dtype=np.float64)
                LHMAT_FSfree = _func(LHMAT=LHMAT, IDX_nFS=IDX_nFS, LHMAT_FSfree=LHMAT_FSfree)

            t8 = time.time()
            # * -- -- -- -- -- -- -- -- Solve Linear System  -- -- -- -- -- -- -- -- *
            if not ConstPhotRatio:
                Solution = np.linalg.solve(LHMAT, RHb).astype(np.float64)

            if ConstPhotRatio:
                Solution_FSfree = np.linalg.solve(LHMAT_FSfree, RHb_FSfree).astype(np.float64)

                # Extend the solution to be consistent form
                _func = SFFTModule_dict['Extend_Solution']
                Solution = np.zeros(NEQ, dtype=np.float64)
                Solution = _func(Solution_FSfree=Solution_FSfree, IDX_nFS=IDX_nFS, Solution=Solution)

            a_ijab = Solution[: Fijab]
            b_pq = Solution[Fijab: ]

            dt8 = time.time() - t8
            dtb = time.time() - tb

            if VERBOSE_LEVEL in [1, 2]:
                print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Establish & Solve Linear System takes [%.4fs]' %dtb)

            if VERBOSE_LEVEL in [2]:
                print('/////   d   ///// Establish OMG                       (%.4fs)' %dt3)
                print('/////   e   ///// Establish GAM                       (%.4fs)' %dt4)
                print('/////   f   ///// Establish PSI                       (%.4fs)' %dt5)
                print('/////   g   ///// Establish PHI                       (%.4fs)' %dt6)
                print('/////   h   ///// Establish THE & DEL                 (%.4fs)' %dt7)
                print('/////   i   ///// Solve Linear System                 (%.4fs)' %dt8)
        
        # * Perform Subtraction
        PixA_DIFF = None
        if Subtract:

            tc = time.time()
            t9 = time.time()
            # Calculate Kab components
            Wl = np.exp((-2j*np.pi/N0) * PixA_X.astype(np.float64))    # row index l, [0, N0)
            Wm = np.exp((-2j*np.pi/N1) * PixA_Y.astype(np.float64))    # column index m, [0, N1)
            Kab_Wla = np.empty((L0, N0, N1), dtype=np.complex128)
            Kab_Wmb = np.empty((L1, N0, N1), dtype=np.complex128)

            if w0 == w1:
                wx = w0   # a little bit faster
                for aob in range(-wx, wx+1):
                    Kab_Wla[aob + wx] = Wl ** aob    # offset 
                    Kab_Wmb[aob + wx] = Wm ** aob    # offset 
            else:
                for a in range(-w0, w0+1): 
                    Kab_Wla[a + w0] = Wl ** a      # offset 
                for b in range(-w1, w1+1): 
                    Kab_Wmb[b + w1] = Wm ** b      # offset 
            dt9 = time.time() - t9

            t10 = time.time()
            # Construct Difference in Fourier Space
            _func = SFFTModule_dict['Construct_FDIFF']   
            PixA_FDIFF = np.empty((N0, N1), dtype=np.complex128)
            PixA_FDIFF = _func(SREF_ijab=SREF_ijab, REF_ab=REF_ab, a_ijab=a_ijab.astype(np.complex128), \
                SPixA_FIij=SPixA_FIij, Kab_Wla=Kab_Wla, Kab_Wmb=Kab_Wmb, b_pq=b_pq.astype(np.complex128), \
                SPixA_FTpq=SPixA_FTpq, PixA_FJ=PixA_FJ, PixA_FDIFF=PixA_FDIFF)

            # Get Difference & Reconstructed Images
            PixA_DIFF = np.empty_like(PixA_FDIFF)
            PixA_DIFF[:, :] = SCALE_L * pyfftw.interfaces.numpy_fft.ifft2(PixA_FDIFF)
            
            PixA_DIFF = PixA_DIFF.real
            dt10 = time.time() - t10
            dtc = time.time() - tc

            if VERBOSE_LEVEL in [1, 2]:
                print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Perform Subtraction takes [%.4fs]' %dtc)
            
            if VERBOSE_LEVEL in [2]:
                print('/////   j   ///// Calculate Kab         (%.4fs)' %dt9)
                print('/////   k   ///// Construct DIFF        (%.4fs)' %dt10)
        
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --||--||--||--||-- EXIT SFFT SUBTRACTION --||--||--||--||-- ')

        return Solution, PixA_DIFF

class ElementalSFFTSubtract:
    @staticmethod
    def ESS(PixA_I, PixA_J, SFFTConfig, SFFTSolution=None, Subtract=False, \
        BACKEND_4SUBTRACT='Cupy', NUM_CPU_THREADS_4SUBTRACT=8, VERBOSE_LEVEL=2):

        if BACKEND_4SUBTRACT == 'Cupy':
            Solution, PixA_DIFF = ElementalSFFTSubtract_Cupy.ESSC(PixA_I=PixA_I, PixA_J=PixA_J, \
                SFFTConfig=SFFTConfig, SFFTSolution=SFFTSolution, Subtract=Subtract, VERBOSE_LEVEL=VERBOSE_LEVEL)

        if BACKEND_4SUBTRACT == 'Numpy':
            Solution, PixA_DIFF = ElementalSFFTSubtract_Numpy.ESSN(PixA_I=PixA_I, PixA_J=PixA_J, \
                SFFTConfig=SFFTConfig, SFFTSolution=SFFTSolution, Subtract=Subtract, \
                NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, VERBOSE_LEVEL=VERBOSE_LEVEL)
        
        return Solution, PixA_DIFF

class GeneralSFFTSubtract:
    @staticmethod
    def GSS(PixA_I, PixA_J, PixA_mI, PixA_mJ, SFFTConfig, ContamMask_I=None, \
        BACKEND_4SUBTRACT='Cupy', NUM_CPU_THREADS_4SUBTRACT=8, VERBOSE_LEVEL=2):
        """
        # Perform image subtraction on I & J with SFFT parameters solved from mI & mJ.
        #
        # Arguments:
        # a) PixA_I: Image I that will be convolved [NaN-Free]
        # b) PixA_J: Image J that won't be convolved [NaN-Free]
        # c) PixA_mI: Masked version of Image I. 'same' means it is identical with I [NaN-Free]
        # d) PixA_mJ: Masked version of Image J. 'same' means it is identical with J [NaN-Free]
        # e) SFFTConfig: Configuration of SFFT
        # f) ContamMask_I: Contaminate-Region in Image I (e.g., Saturation and Bad pixels).
        #                  We will calculate the propagated Contaminate-Region on convolved I.
        # 
        # g) BACKEND_4SUBTRACT: Which backend would you like to perform SFFT subtraction on? (Cupy/Numpy)
        # i) NUM_CPU_THREADS_4SUBTRACT: How many CPU threads would you want to perform SFFT subtraction on? (Numpy)
        #
        """
        
        SFFT_BANNER = r"""
                                __    __    __    __
                               /  \  /  \  /  \  /  \
                              /    \/    \/    \/    \
            █████████████████/  /██/  /██/  /██/  /█████████████████████████
                            /  / \   / \   / \   / \  \____
                           /  /   \_/   \_/   \_/   \    o \__,
                          / _/                       \_____/  `
                          |/
        
                      █████████  ███████████ ███████████ ███████████        
                     ███░░░░░███░░███░░░░░░█░░███░░░░░░█░█░░░███░░░█            
                    ░███    ░░░  ░███   █ ░  ░███   █ ░ ░   ░███  ░ 
                    ░░█████████  ░███████    ░███████       ░███    
                     ░░░░░░░░███ ░███░░░█    ░███░░░█       ░███    
                     ███    ░███ ░███  ░     ░███  ░        ░███    
                    ░░█████████  █████       █████          █████   
                     ░░░░░░░░░  ░░░░░       ░░░░░          ░░░░░         
        
                    Saccadic Fast Fourier Transform (SFFT) algorithm
                    sfft supported by @LeiHu
        
                    GitHub: https://github.com/thomasvrussell/sfft
                    Related Paper: https://arxiv.org/abs/2109.09334
                    
            ████████████████████████████████████████████████████████████████
            
            """
        if VERBOSE_LEVEL in [2]:
            print(SFFT_BANNER)
        
        # * Size-Check
        tmplst = [PixA_I.shape, PixA_J.shape, PixA_mI.shape, PixA_mI.shape]
        if len(set(tmplst)) > 1:
            raise Exception('MeLOn ERROR: Input images should have same size!')
        
        # * Subtraction Solution derived from input masked image-pair
        Solution = ElementalSFFTSubtract.ESS(PixA_I=PixA_mI, PixA_J=PixA_mJ, \
            SFFTConfig=SFFTConfig, SFFTSolution=None, Subtract=False, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
            NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, VERBOSE_LEVEL=VERBOSE_LEVEL)[0]
        
        # * Subtraction of the input image-pair (use above solution)
        PixA_DIFF = ElementalSFFTSubtract.ESS(PixA_I=PixA_I, PixA_J=PixA_J, \
            SFFTConfig=SFFTConfig, SFFTSolution=Solution, Subtract=True, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
            NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, VERBOSE_LEVEL=VERBOSE_LEVEL)[1]
        
        # * Identify propagated contamination region through convolving I
        ContamMask_CI = None
        if ContamMask_I is not None:
            tSolution = Solution.copy()
            DB = SFFTConfig[0]['DB']
            Fpq = int((DB+1)*(DB+2)/2)
            tSolution[-Fpq:] = 0.0

            _tmpI = ContamMask_I.astype(np.float64)
            _tmpJ = np.zeros(PixA_J.shape).astype(np.float64)
            _tmpD = ElementalSFFTSubtract.ESS(PixA_I=_tmpI, PixA_J=_tmpJ, \
                SFFTConfig=SFFTConfig, SFFTSolution=tSolution, Subtract=True, BACKEND_4SUBTRACT=BACKEND_4SUBTRACT, \
                NUM_CPU_THREADS_4SUBTRACT=NUM_CPU_THREADS_4SUBTRACT, VERBOSE_LEVEL=VERBOSE_LEVEL)[1]
            
            FTHRESH = -0.001  # Emperical
            ContamMask_CI = _tmpD < FTHRESH
        
        return Solution, PixA_DIFF, ContamMask_CI

# add a version with GPU arrays as inputs & outputs
class ElementalSFFTSubtract_PureCupy:
    @staticmethod
    def ESSPC(PixA_I_GPU, PixA_J_GPU, SFFTConfig, SFFTSolution_GPU=None, Subtract=False, VERBOSE_LEVEL=2):
        
        import cupy as cp
        import cupyx.scipy.linalg as cpx_linalg

        def LSSolver(LHMAT_GPU, RHb_GPU):
            """
            # NOTE: cupy/numpy (moderately/extremely) slow version has been DEPRECATED!
            Solution_GPU = cp.linalg.solve(LHMAT_GPU, RHb_GPU)
            Solution_GPU = cp.array(np.linalg.solve(cp.asnumpy(LHMAT_GPU), cp.asnumpy(RHb_GPU)), dtype=np.float64)
            """
            lu_piv_GPU = cpx_linalg.lu_factor(LHMAT_GPU, overwrite_a=False, check_finite=True)
            Solution_GPU = cpx_linalg.lu_solve(lu_piv_GPU, RHb_GPU)
            return Solution_GPU
        
        ta = time.time()
        # * Read SFFT parameters
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --||--||--||--||-- TRIGGER SFFT SUBTRACTION --||--||--||--||-- ')
            print('\n ---||--- KerPolyOrder %d | BGPolyOrder %d | KerHW [%d] ---||--- '\
                   %(SFFTConfig[0]['DK'], SFFTConfig[0]['DB'], SFFTConfig[0]['w0']))

        SFFTParam_dict, SFFTModule_dict = SFFTConfig
        N0, N1 = SFFTParam_dict['N0'], SFFTParam_dict['N1']
        w0, w1 = SFFTParam_dict['w0'], SFFTParam_dict['w1']
        DK, DB = SFFTParam_dict['DK'], SFFTParam_dict['DB']

        if PixA_I_GPU.shape != (N0, N1) or PixA_J_GPU.shape != (N0, N1):
            _error_message = 'INCONSISTENT shape of input images I & J, [%d, %d] required!' %(N0, N1)
            raise Exception('MeLOn ERROR: %s' %_error_message)

        ConstPhotRatio = SFFTParam_dict['ConstPhotRatio']
        MaxThreadPerB = SFFTParam_dict['MaxThreadPerB']
        
        L0, L1 = SFFTParam_dict['L0'], SFFTParam_dict['L1']
        Fab, Fij, Fpq = SFFTParam_dict['Fab'], SFFTParam_dict['Fij'], SFFTParam_dict['Fpq']
        SCALE, SCALE_L = SFFTParam_dict['SCALE'], SFFTParam_dict['SCALE_L']

        NEQ, Fijab = SFFTParam_dict['NEQ'], SFFTParam_dict['Fijab']
        NEQ_FSfree = SFFTParam_dict['NEQ_FSfree']

        FOMG, FGAM = SFFTParam_dict['FOMG'], SFFTParam_dict['FGAM']
        FTHE, FPSI = SFFTParam_dict['FTHE'], SFFTParam_dict['FPSI']
        FPHI, FDEL = SFFTParam_dict['FPHI'], SFFTParam_dict['FDEL']
        
        # * Grid-Block-Thread Managaement [Pixel-Level]
        GPUManage = lambda NT: ((NT-1)//MaxThreadPerB + 1, min(NT, MaxThreadPerB))
        BpG_PIX0, TpB_PIX0 = GPUManage(N0)
        BpG_PIX1, TpB_PIX1 = GPUManage(N1)
        BpG_PIX, TpB_PIX = (BpG_PIX0, BpG_PIX1), (TpB_PIX0, TpB_PIX1, 1)

        # * Define First-order MultiIndex Reference
        REF_pq_GPU = cp.array([(p, q) for p in range(DB+1) for q in range(DB+1-p)], dtype=cp.int32)
        REF_ij_GPU = cp.array([(i, j) for i in range(DK+1) for j in range(DK+1-i)], dtype=cp.int32)
        REF_ab_GPU = cp.array([(a_pos-w0, b_pos-w1) for a_pos in range(L0) for b_pos in range(L1)], dtype=cp.int32)

        # * Define Second-order MultiIndex Reference
        SREF_iji0j0_GPU = cp.array([(ij, i0j0) for ij in range(Fij) for i0j0 in range(Fij)], dtype=cp.int32)
        SREF_pqp0q0_GPU = cp.array([(pq, p0q0) for pq in range(Fpq) for p0q0 in range(Fpq)], dtype=cp.int32)
        SREF_ijpq_GPU = cp.array([(ij, pq) for ij in range(Fij) for pq in range(Fpq)], dtype=cp.int32)
        SREF_pqij_GPU = cp.array([(pq, ij) for pq in range(Fpq) for ij in range(Fij)], dtype=cp.int32)
        SREF_ijab_GPU = cp.array([(ij, ab) for ij in range(Fij) for ab in range(Fab)], dtype=cp.int32)

        # * Define the Forbidden ij-ab Rule
        ij00 = np.arange(w0 * L1 + w1, Fijab, Fab).astype(np.int32)
        if ConstPhotRatio:
            FBijab = ij00[1:]
            MASK_nFS = np.ones(NEQ).astype(bool)
            MASK_nFS[FBijab] = False            
            IDX_nFS = np.where(MASK_nFS)[0].astype(np.int32)
            IDX_nFS_GPU = cp.array(IDX_nFS)
            NEQ_FSfree = len(IDX_nFS)

        t0 = time.time()
        # * Read input images as C-order arrays
        assert PixA_I_GPU.flags['C_CONTIGUOUS']
        assert PixA_J_GPU.flags['C_CONTIGUOUS']
        dt0 = time.time() - t0

        # * Symbol Convention NOTE
        #    X (x) / Y (y) ----- pixel row / column index
        #    CX (cx) / CY (cy) ----- ScaledFortranCoor of pixel (x, y) center   
        #    e.g. pixel (x, y) = (3, 5) corresponds (cx, cy) = (4.0/N0, 6.0/N1)
        #    NOTE cx / cy is literally \mathtt{x} / \mathtt{y} in sfft paper.
        #    NOTE Without special definition, MeLOn convention refers to X (x) / Y (y) as FortranCoor.

        # * Get Spatial Coordinates
        t1 = time.time()
        PixA_X_GPU = cp.zeros((N0, N1), dtype=cp.int32)      # row index, [0, N0)
        PixA_Y_GPU = cp.zeros((N0, N1), dtype=cp.int32)      # column index, [0, N1)
        PixA_CX_GPU = cp.zeros((N0, N1), dtype=cp.float64)   # coordinate.x
        PixA_CY_GPU = cp.zeros((N0, N1), dtype=cp.float64)   # coordinate.y 

        _module = SFFTModule_dict['SpatialCoor']
        _func = _module.get_function('kmain')
        _func(args=(PixA_X_GPU, PixA_Y_GPU, PixA_CX_GPU, PixA_CY_GPU), block=TpB_PIX, grid=BpG_PIX)

        # * Spatial Polynomial terms Iij, Tpq
        SPixA_Iij_GPU = cp.zeros((Fij, N0, N1), dtype=cp.float64)
        SPixA_Tpq_GPU = cp.zeros((Fpq, N0, N1), dtype=cp.float64)

        _module = SFFTModule_dict['SpatialPoly']
        _func = _module.get_function('kmain')
        _func(args=(REF_ij_GPU, REF_pq_GPU, PixA_CX_GPU, PixA_CY_GPU, PixA_I_GPU, \
            SPixA_Iij_GPU, SPixA_Tpq_GPU), block=TpB_PIX, grid=BpG_PIX)
        
        # del PixA_I_GPU
        dt1 = time.time() - t1

        t2 = time.time()

        # * Make DFT of J, Iij, Tpq and their conjugates
        PixA_FJ_GPU = cp.empty((N0, N1), dtype=cp.complex128)
        PixA_FJ_GPU[:, :] = PixA_J_GPU.astype(cp.complex128)
        PixA_FJ_GPU[:, :] = cp.fft.fft2(PixA_FJ_GPU)
        PixA_FJ_GPU[:, :] *= SCALE

        SPixA_FIij_GPU = cp.empty((Fij, N0, N1), dtype=cp.complex128)
        SPixA_FIij_GPU[:, :, :] = SPixA_Iij_GPU.astype(cp.complex128)
        for k in range(Fij): 
            SPixA_FIij_GPU[k: k+1] = cp.fft.fft2(SPixA_FIij_GPU[k: k+1])
        SPixA_FIij_GPU[:, :] *= SCALE

        SPixA_FTpq_GPU = cp.empty((Fpq, N0, N1), dtype=cp.complex128)
        SPixA_FTpq_GPU[:, :, :] = SPixA_Tpq_GPU.astype(cp.complex128)
        for k in range(Fpq): 
            SPixA_FTpq_GPU[k: k+1] = cp.fft.fft2(SPixA_FTpq_GPU[k: k+1])
        SPixA_FTpq_GPU[:, :] *= SCALE

        del SPixA_Iij_GPU
        del SPixA_Tpq_GPU

        PixA_CFJ_GPU = cp.conj(PixA_FJ_GPU)
        SPixA_CFIij_GPU = cp.conj(SPixA_FIij_GPU)
        SPixA_CFTpq_GPU = cp.conj(SPixA_FTpq_GPU)
        dt2 = time.time() - t2
        dta = time.time() - ta

        if VERBOSE_LEVEL in [1, 2]:
            print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Preliminary Steps takes [%.4fs]' %dta)

        if VERBOSE_LEVEL in [2]:
            print('/////   a   ///// Read Input Images  (%.4fs)' %dt0)
            print('/////   b   ///// Spatial Polynomial (%.4fs)' %dt1)
            print('/////   c   ///// DFT-%d             (%.4fs)' %(1 + Fij + Fpq, dt2))

        # * Consider The Major Sources of the Linear System 
        #     𝛀_i8j8ij     &    𝜦_i8j8pq    ||     𝚯_i8j8
        #     𝜳_p8q8ij     &    𝚽_p8q8pq    ||     𝚫_p8q8   
        #
        # * Remarks
        #    a. They have consistent form: Greek(rho, eps) = PreGreek(Mod_N0(rho), Mod_N1(eps))
        #    b. PreGreek = s * Re[DFT(HpGreek)], where HpGreek = GLH * GRH and s is some real-scale.
        #    c. Considering the subscripted variables, HpGreek / PreGreek is Complex / Real 3D with shape (F_Greek, N0, N1).
        
        if SFFTSolution_GPU is not None:
            assert SFFTSolution_GPU.dtype == cp.float64
            Solution_GPU = SFFTSolution_GPU
            a_ijab_GPU = Solution_GPU[: Fijab]
            b_pq_GPU = Solution_GPU[Fijab: ]

        if SFFTSolution_GPU is None:
            tb = time.time()
            # * Grid-Block-Thread Managaement [for Greek]
            BpG_OMG0, TpB_OMG0 = GPUManage(Fijab)
            BpG_OMG1, TpB_OMG1 = GPUManage(Fijab)
            BpG_OMG, TpB_OMG = (BpG_OMG0, BpG_OMG1), (TpB_OMG0, TpB_OMG1, 1)

            BpG_GAM0, TpB_GAM0 = GPUManage(Fijab)
            BpG_GAM1, TpB_GAM1 = GPUManage(Fpq)
            BpG_GAM, TpB_GAM = (BpG_GAM0, BpG_GAM1), (TpB_GAM0, TpB_GAM1, 1)

            BpG_THE0, TpB_THE0 = GPUManage(Fijab)
            BpG_THE, TpB_THE = (BpG_THE0, 1), (TpB_THE0, 1, 1)

            BpG_PSI0, TpB_PSI0 = GPUManage(Fpq)
            BpG_PSI1, TpB_PSI1 = GPUManage(Fijab)
            BpG_PSI, TpB_PSI = (BpG_PSI0, BpG_PSI1), (TpB_PSI0, TpB_PSI1, 1)

            BpG_PHI0, TpB_PHI0 = GPUManage(Fpq)
            BpG_PHI1, TpB_PHI1 = GPUManage(Fpq)
            BpG_PHI, TpB_PHI = (BpG_PHI0, BpG_PHI1), (TpB_PHI0, TpB_PHI1, 1)

            BpG_DEL0, TpB_DEL0 = GPUManage(Fpq)
            BpG_DEL, TpB_DEL = (BpG_DEL0, 1), (TpB_DEL0, 1, 1)

            LHMAT_GPU = cp.empty((NEQ, NEQ), dtype=cp.float64)
            RHb_GPU = cp.empty(NEQ, dtype=cp.float64)
            t3 = time.time()

            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝛀  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝛀 [HpOMG]
            _module = SFFTModule_dict['HadProd_OMG']
            _func = _module.get_function('kmain')
            HpOMG_GPU = cp.empty((FOMG, N0, N1), dtype=cp.complex128)
            _func(args=(SREF_iji0j0_GPU, SPixA_FIij_GPU, SPixA_CFIij_GPU, HpOMG_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PreOMG = SCALE * Re[DFT(HpOMG)]
            for k in range(FOMG):
                HpOMG_GPU[k: k+1] = cp.fft.fft2(HpOMG_GPU[k: k+1])
            HpOMG_GPU *= SCALE

            PreOMG_GPU = cp.empty((FOMG, N0, N1), dtype=cp.float64)
            PreOMG_GPU[:, :, :] = HpOMG_GPU.real
            PreOMG_GPU[:, :, :] *= SCALE
            del HpOMG_GPU

            # c. Fill Linear System with PreOMG
            _module = SFFTModule_dict['FillLS_OMG']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PreOMG_GPU, LHMAT_GPU), \
                block=TpB_OMG, grid=BpG_OMG)

            del PreOMG_GPU
            dt3 = time.time() - t3

            t4 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝜦  -- -- -- -- -- -- -- -- *

            # a. Hadamard Product for 𝜦 [HpGAM]
            _module = SFFTModule_dict['HadProd_GAM']
            _func = _module.get_function('kmain')
            HpGAM_GPU = cp.empty((FGAM, N0, N1), dtype=cp.complex128)
            _func(args=(SREF_ijpq_GPU, SPixA_FIij_GPU, SPixA_CFTpq_GPU, HpGAM_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PreGAM = 1 * Re[DFT(HpGAM)]
            for k in range(FGAM):
                HpGAM_GPU[k: k+1] = cp.fft.fft2(HpGAM_GPU[k: k+1])
            HpGAM_GPU *= SCALE

            PreGAM_GPU = cp.empty((FGAM, N0, N1), dtype=cp.float64)
            PreGAM_GPU[:, :, :] = HpGAM_GPU.real
            del HpGAM_GPU

            # c. Fill Linear System with PreGAM
            _module = SFFTModule_dict['FillLS_GAM']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PreGAM_GPU, LHMAT_GPU), \
                block=TpB_GAM, grid=BpG_GAM)

            del PreGAM_GPU
            dt4 = time.time() - t4

            t5 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝜳  -- -- -- -- -- -- -- -- *
            
            # a. Hadamard Product for 𝜳  [HpPSI]
            _module = SFFTModule_dict['HadProd_PSI']
            _func = _module.get_function('kmain')
            HpPSI_GPU = cp.empty((FPSI, N0, N1), dtype=cp.complex128)
            _func(args=(SREF_pqij_GPU, SPixA_CFIij_GPU, SPixA_FTpq_GPU, HpPSI_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PrePSI = 1 * Re[DFT(HpPSI)]
            for k in range(FPSI):
                HpPSI_GPU[k: k+1] = cp.fft.fft2(HpPSI_GPU[k: k+1])
            HpPSI_GPU *= SCALE

            PrePSI_GPU = cp.empty((FPSI, N0, N1), dtype=cp.float64)
            PrePSI_GPU[:, :, :] = HpPSI_GPU.real
            del HpPSI_GPU

            # c. Fill Linear System with PrePSI
            _module = SFFTModule_dict['FillLS_PSI']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PrePSI_GPU, LHMAT_GPU), \
                block=TpB_PSI, grid=BpG_PSI)

            del PrePSI_GPU
            dt5 = time.time() - t5

            t6 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝚽  -- -- -- -- -- -- -- -- *
            
            # a. Hadamard Product for 𝚽  [HpPHI]
            _module = SFFTModule_dict['HadProd_PHI']
            _func = _module.get_function('kmain')
            HpPHI_GPU = cp.empty((FPHI, N0, N1), dtype=cp.complex128)
            _func(args=(SREF_pqp0q0_GPU, SPixA_FTpq_GPU, SPixA_CFTpq_GPU, HpPHI_GPU), \
                block=TpB_PIX, grid=BpG_PIX)

            # b. PrePHI = SCALE_L * Re[DFT(HpPHI)]
            for k in range(FPHI):
                HpPHI_GPU[k: k+1] = cp.fft.fft2(HpPHI_GPU[k: k+1])
            HpPHI_GPU *= SCALE

            PrePHI_GPU = cp.empty((FPHI, N0, N1), dtype=cp.float64)
            PrePHI_GPU[:, :, :] = HpPHI_GPU.real
            PrePHI_GPU[:, :, :] *= SCALE_L
            del HpPHI_GPU

            # c. Fill Linear System with PrePHI
            _module = SFFTModule_dict['FillLS_PHI']
            _func = _module.get_function('kmain')
            _func(args=(PrePHI_GPU, LHMAT_GPU), block=TpB_PHI, grid=BpG_PHI)

            del PrePHI_GPU
            dt6 = time.time() - t6

            t7 = time.time()
            # * -- -- -- -- -- -- -- -- Establish Linear System through 𝚯 & 𝚫  -- -- -- -- -- -- -- -- *

            # a1. Hadamard Product for 𝚯  [HpTHE]
            _module = SFFTModule_dict['HadProd_THE']
            _func = _module.get_function('kmain')
            HpTHE_GPU = cp.empty((FTHE, N0, N1), dtype=cp.complex128)
            _func(args=(SPixA_FIij_GPU, PixA_CFJ_GPU, HpTHE_GPU), block=TpB_PIX, grid=BpG_PIX)

            # a2. Hadamard Product for 𝚫  [HpDEL]
            _module = SFFTModule_dict['HadProd_DEL']
            _func = _module.get_function('kmain')
            HpDEL_GPU = cp.empty((FDEL, N0, N1), dtype=cp.complex128)
            _func(args=(SPixA_FTpq_GPU, PixA_CFJ_GPU, HpDEL_GPU), block=TpB_PIX, grid=BpG_PIX)

            # b1. PreTHE = 1 * Re[DFT(HpTHE)]
            # b2. PreDEL = SCALE_L * Re[DFT(HpDEL)]
            for k in range(FTHE):
                HpTHE_GPU[k: k+1] = cp.fft.fft2(HpTHE_GPU[k: k+1])
            HpTHE_GPU[:, :, :] *= SCALE
            
            for k in range(FDEL):
                HpDEL_GPU[k: k+1] = cp.fft.fft2(HpDEL_GPU[k: k+1])
            HpDEL_GPU[:, :, :] *= SCALE

            PreTHE_GPU = cp.empty((FTHE, N0, N1), dtype=cp.float64)
            PreTHE_GPU[:, :, :] = HpTHE_GPU.real
            del HpTHE_GPU
            
            PreDEL_GPU = cp.empty((FDEL, N0, N1), dtype=cp.float64)
            PreDEL_GPU[:, :, :] = HpDEL_GPU.real
            PreDEL_GPU[:, :, :] *= SCALE_L
            del HpDEL_GPU
            
            # c1. Fill Linear System with PreTHE
            _module = SFFTModule_dict['FillLS_THE']
            _func = _module.get_function('kmain')
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, PreTHE_GPU, RHb_GPU), \
                block=TpB_THE, grid=BpG_THE)
            del PreTHE_GPU

            # c2. Fill Linear System with PreDEL
            _module = SFFTModule_dict['FillLS_DEL']
            _func = _module.get_function('kmain')
            _func(args=(PreDEL_GPU, RHb_GPU), block=TpB_DEL, grid=BpG_DEL)
            
            del PreDEL_GPU
            dt7 = time.time() - t7

            # * -- -- -- -- -- -- -- -- Remove Forbidden Stripes  -- -- -- -- -- -- -- -- *
            if not ConstPhotRatio: pass
            if ConstPhotRatio:
                RHb_FSfree_GPU = cp.take(RHb_GPU, IDX_nFS_GPU)
                _module = SFFTModule_dict['Remove_LSFStripes']
                _func = _module.get_function('kmain')
                BpG_FSfree_PA, TpB_FSfree_PA = GPUManage(NEQ_FSfree)     # Per Axis
                BpG_FSfree, TpB_FSfree = (BpG_FSfree_PA, BpG_FSfree_PA), (TpB_FSfree_PA, TpB_FSfree_PA, 1)
                LHMAT_FSfree_GPU = cp.empty((NEQ_FSfree, NEQ_FSfree), dtype=cp.float64)
                _func(args=(LHMAT_GPU, IDX_nFS_GPU, LHMAT_FSfree_GPU), block=TpB_FSfree, grid=BpG_FSfree)

            t8 = time.time()
            # * -- -- -- -- -- -- -- -- Solve Linear System  -- -- -- -- -- -- -- -- *
            if not ConstPhotRatio: 
                Solution_GPU = LSSolver(LHMAT_GPU=LHMAT_GPU, RHb_GPU=RHb_GPU)

            if ConstPhotRatio:
                # Extend the solution to be consistent form
                Solution_FSfree_GPU = LSSolver(LHMAT_GPU=LHMAT_FSfree_GPU, RHb_GPU=RHb_FSfree_GPU)
                _module = SFFTModule_dict['Extend_Solution']
                _func = _module.get_function('kmain')
                BpG_ES, TpB_ES = (BpG_FSfree_PA, 1), (TpB_FSfree_PA, 1, 1)
                Solution_GPU = cp.zeros(NEQ, dtype=cp.float64)
                _func(args=(Solution_FSfree_GPU, IDX_nFS_GPU, Solution_GPU), block=TpB_ES, grid=BpG_ES)
            
            a_ijab_GPU = Solution_GPU[: Fijab]
            b_pq_GPU = Solution_GPU[Fijab: ]
            
            dt8 = time.time() - t8
            dtb = time.time() - tb

            if VERBOSE_LEVEL in [1, 2]:
                print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Establish & Solve Linear System takes [%.4fs]' %dtb)

            if VERBOSE_LEVEL in [2]:
                print('/////   d   ///// Establish OMG                       (%.4fs)' %dt3)
                print('/////   e   ///// Establish GAM                       (%.4fs)' %dt4)
                print('/////   f   ///// Establish PSI                       (%.4fs)' %dt5)
                print('/////   g   ///// Establish PHI                       (%.4fs)' %dt6)
                print('/////   h   ///// Establish THE & DEL                 (%.4fs)' %dt7)
                print('/////   i   ///// Solve Linear System                 (%.4fs)' %dt8)

        # * Perform Subtraction 
        PixA_DIFF_GPU = None
        if Subtract:
            tc = time.time()
            t9 = time.time()
            # Calculate Kab components
            Wl_GPU = cp.exp((-2j*cp.pi/N0) * PixA_X_GPU.astype(cp.float64))    # row index l, [0, N0)
            Wm_GPU = cp.exp((-2j*cp.pi/N1) * PixA_Y_GPU.astype(cp.float64))    # column index m, [0, N1)
            Kab_Wla_GPU = cp.empty((L0, N0, N1), dtype=cp.complex128)
            Kab_Wmb_GPU = cp.empty((L1, N0, N1), dtype=cp.complex128)

            if w0 == w1:
                wx = w0   # a little bit faster
                for aob in range(-wx, wx+1):
                    Kab_Wla_GPU[aob + wx] = Wl_GPU ** aob    # offset 
                    Kab_Wmb_GPU[aob + wx] = Wm_GPU ** aob    # offset 
            else:
                for a in range(-w0, w0+1): 
                    Kab_Wla_GPU[a + w0] = Wl_GPU ** a        # offset 
                for b in range(-w1, w1+1): 
                    Kab_Wmb_GPU[b + w1] = Wm_GPU ** b        # offset 
            dt9 = time.time() - t9

            t10 = time.time()
            # Construct Difference in Fourier Space
            _module = SFFTModule_dict['Construct_FDIFF']
            _func = _module.get_function('kmain')     
            PixA_FDIFF_GPU = cp.empty((N0, N1), dtype=cp.complex128)
            _func(args=(SREF_ijab_GPU, REF_ab_GPU, a_ijab_GPU.astype(cp.complex128), \
                SPixA_FIij_GPU, Kab_Wla_GPU, Kab_Wmb_GPU, b_pq_GPU.astype(cp.complex128), \
                SPixA_FTpq_GPU, PixA_FJ_GPU, PixA_FDIFF_GPU), block=TpB_PIX, grid=BpG_PIX)
            
            # Get Difference & Reconstructed Images
            PixA_DIFF_GPU = SCALE_L * cp.fft.ifft2(PixA_FDIFF_GPU).real
            
            dt10 = time.time() - t10
            dtc = time.time() - tc

            if VERBOSE_LEVEL in [1, 2]:
                print('\nMeLOn CheckPoint: SFFT-SUBTRACTION Perform Subtraction takes [%.4fs]' %dtc)
            
            if VERBOSE_LEVEL in [2]:
                print('/////   j   ///// Calculate Kab         (%.4fs)' %dt9)
                print('/////   k   ///// Construct DIFF        (%.4fs)' %dt10)
        
        if VERBOSE_LEVEL in [1, 2]:
            print('\n --||--||--||--||-- EXIT SFFT SUBTRACTION --||--||--||--||-- ')

        return Solution_GPU, PixA_DIFF_GPU

class GeneralSFFTSubtract_PureCupy:
    @staticmethod
    def GSS(PixA_I_GPU, PixA_J_GPU, PixA_mI_GPU, PixA_mJ_GPU, SFFTConfig, ContamMask_I_GPU=None, VERBOSE_LEVEL=2):
        """
        # Perform image subtraction on I & J with SFFT parameters solved from mI & mJ ().
        #
        # Arguments:
        # a) PixA_I_GPU: Image I that will be convolved [NaN-Free]
        # b) PixA_J_GPU: Image J that won't be convolved [NaN-Free]
        # c) PixA_mI_GPU: Masked version of Image I. 'same' means it is identical with I [NaN-Free]
        # d) PixA_mJ_GPU: Masked version of Image J. 'same' means it is identical with J [NaN-Free]
        # e) SFFTConfig: Configuration of SFFT
        # f) ContamMask_I_GPU: Contaminate-Region in Image I (e.g., Saturation and Bad pixels).
        #                      We will calculate the propagated Contaminate-Region on convolved I.
        #
        """
        SFFT_BANNER = r"""
                                __    __    __    __
                               /  \  /  \  /  \  /  \
                              /    \/    \/    \/    \
            █████████████████/  /██/  /██/  /██/  /█████████████████████████
                            /  / \   / \   / \   / \  \____
                           /  /   \_/   \_/   \_/   \    o \__,
                          / _/                       \_____/  `
                          |/
        
                      █████████  ███████████ ███████████ ███████████        
                     ███░░░░░███░░███░░░░░░█░░███░░░░░░█░█░░░███░░░█            
                    ░███    ░░░  ░███   █ ░  ░███   █ ░ ░   ░███  ░ 
                    ░░█████████  ░███████    ░███████       ░███    
                     ░░░░░░░░███ ░███░░░█    ░███░░░█       ░███    
                     ███    ░███ ░███  ░     ░███  ░        ░███    
                    ░░█████████  █████       █████          █████   
                     ░░░░░░░░░  ░░░░░       ░░░░░          ░░░░░         
        
                    Saccadic Fast Fourier Transform (SFFT) algorithm
                    sfft supported by @LeiHu
        
                    GitHub: https://github.com/thomasvrussell/sfft
                    Related Paper: https://arxiv.org/abs/2109.09334
                    
            ████████████████████████████████████████████████████████████████
            
            """
        
        if VERBOSE_LEVEL in [2]:
            print(SFFT_BANNER)
        
        # * Size-Check
        tmplst = [PixA_I_GPU.shape, PixA_J_GPU.shape, PixA_mI_GPU.shape, PixA_mI_GPU.shape]
        if len(set(tmplst)) > 1:
            raise Exception('MeLOn ERROR: Input images should have same size!')
        
        # * Subtraction Solution derived from input masked image-pair
        Solution_GPU = ElementalSFFTSubtract_PureCupy.ESSPC(PixA_I_GPU=PixA_mI_GPU, PixA_J_GPU=PixA_mJ_GPU, 
            SFFTConfig=SFFTConfig, SFFTSolution_GPU=None, Subtract=False, VERBOSE_LEVEL=VERBOSE_LEVEL)[0]
        
        # * Subtraction of the input image-pair (use the solution above)
        PixA_DIFF_GPU = ElementalSFFTSubtract_PureCupy.ESSPC(PixA_I_GPU=PixA_I_GPU, PixA_J_GPU=PixA_J_GPU,
            SFFTConfig=SFFTConfig, SFFTSolution_GPU=Solution_GPU, Subtract=True, VERBOSE_LEVEL=VERBOSE_LEVEL)[1]

        # * Identify propagated contamination region through convolving I
        ContamMask_CI_GPU = None
        if ContamMask_I_GPU is not None:
            import cupy as cp
            DB = SFFTConfig[0]['DB']
            Fpq = int((DB+1)*(DB+2)/2)
            
            _Solution_GPU = Solution_GPU.copy()
            _Solution_GPU[-Fpq:] = 0.
            _PixA_I_GPU = ContamMask_I_GPU.astype(cp.float64)
            _PixA_J_GPU = cp.zeros_like(PixA_J_GPU.shape)
            
            _PixA_DIFF_GPU = ElementalSFFTSubtract_PureCupy.ESSPC(PixA_I_GPU=_PixA_I_GPU, PixA_J_GPU=_PixA_J_GPU,
                SFFTConfig=SFFTConfig, SFFTSolution_GPU=_Solution_GPU, Subtract=True, VERBOSE_LEVEL=VERBOSE_LEVEL)[1]

            FTHRESH = -0.001  # Emperical
            ContamMask_CI_GPU = _PixA_DIFF_GPU < FTHRESH
        
        return Solution_GPU, PixA_DIFF_GPU, ContamMask_CI_GPU



import signal
import threading, time, ctypes

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.1"

class Timeout():
    """ Timeout class using ALARM signal """
    class Timeout(Exception):
        pass
    def __init__(self, sec):
        self.sec = sec
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.raise_timeout)
        signal.alarm(self.sec)
    def __exit__(self, *args):
        signal.alarm(0)  # disable alarm
    def raise_timeout(self, *args):
        raise Timeout.Timeout()

class TimeoutAfter():
	def __init__(self, timeout=(10), exception=TimeoutError):

        # ************************* Important NOTE ************************* #
        # the function is from https://github.com/dr-luke/PyTimeoutAfter,
        # which allows for raising timeout error in a child thread.
        #
        # -----------------------------------------------------------------------
        # Copyright (c) 2021 Levi M. Luke, LIU Wei, Brett Husar, and others
        # -----------------------------------------------------------------------
        #
        # ************************* Important NOTE ************************* #

		self._exception = exception
		self._caller_thread = threading.current_thread()
		self._timeout = timeout
		self._timer = threading.Timer(self._timeout, self.raise_caller)
		self._timer.daemon = True
		self._timer.start()

	def __enter__(self):
		try:
			yield
		finally:
			self._timer.cancel()
		return self

	def __exit__(self, type, value, traceback):
		self._timer.cancel()
		
	def raise_caller(self):
		ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._caller_thread._ident), ctypes.py_object(self._exception))
		if ret == 0:
			raise ValueError("Invalid thread ID")
		elif ret > 1:
			ctypes.pythonapi.PyThreadState_SetAsyncExc(self._caller_thread._ident, NULL)
			raise SystemError("PyThreadState_SetAsyncExc failed")
            


import time
import numpy as np
import importlib.util as iu
from sfft.sfftcore.SFFTSubtract import GeneralSFFTSubtract_PureCupy
from sfft.sfftcore.SFFTConfigure import SingleSFFTConfigure
if iu.find_spec('cupy') is not None:
    import cupy as cp

__last_update__ = "2024-10-17"
__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.6.3"

#########################################################################
#||                                                                   ||#
#||                           Acknowledgment                          ||#
#||                                                                   ||#
#########################################################################
#||                                                                   ||#
#||   This code was developed during the NASA GPU Hackathon 2024      ||#
#||   (https://www.nas.nasa.gov/hackathon/#home) as part of the       ||#
#||   pipeline optimization efforts for the Roman Supernova PIT       ||#
#||   team (https://github.com/Roman-Supernova-PIT). I extend my      ||#
#||   sincere appreciation to my team members and our Hackathon       ||#
#||   mentors for their valuable support and insightful               ||#
#||   suggestions.                                                    ||#
#||                                                                   ||#
#||   I would like to specifically acknowledge the contributions      ||#
#||   of the following team members: Lauren Aldoroty (Duke),          ||#
#||   Robert Knop (LBNL), Shu Liu (Pitt), and Michael Wood-Vasey      ||#
#||   (Pitt). Additionally, I am grateful for the guidance provided   ||#
#||   by our mentors, Marcus Manos and Lucas Erlandson from NVIDIA.   ||#
#||                                                                   ||#
#||   The collaborative environment fostered by the NASA GPU          ||#
#||   Hackathon and the Roman Supernova PIT team was essential for    ||#
#||   the development and optimization of this code.                  ||#
#||                                                                   ||#
#########################################################################

class PureCupy_Customized_Packet:
    @staticmethod
    def PCCP(PixA_REF_GPU, PixA_SCI_GPU, PixA_mREF_GPU, PixA_mSCI_GPU, 
        ForceConv, GKerHW, KerPolyOrder=2, BGPolyOrder=2, ConstPhotRatio=True, \
        CUDA_DEVICE_4SUBTRACT='0', VERBOSE_LEVEL=2):
        """
        * Parameters for Cupy Customized SFFT (Pure Cupy Version)
        # @ Customized: Skip built-in preprocessing (automatic image-masking) by a customized masked image-pair.
        #
        # ----------------------------- Computing Enviornment --------------------------------- #

        -CUDA_DEVICE_4SUBTRACT ['0']        # it specifies certain GPU device (index) to conduct the subtraction task.
                                            # the GPU devices are usually numbered 0 to N-1 (you may use command nvidia-smi to check).
                                            # this argument becomes trivial for Numpy backend.

        # ----------------------------- SFFT Subtraction --------------------------------- #

        -ForceConv []                       # it determines which image will be convolved, can be 'REF' or 'SCI'.
                                            # here ForceConv CANNOT be 'AUTO'!

        -GKerHW []                          # the given kernel half-width. 

        -KerPolyOrder [2]                   # Polynomial degree of kernel spatial variation.

        -BGPolyOrder [2]                    # Polynomial degree of background spatial variation.

        -ConstPhotRatio [True]              # Constant photometric ratio between images ? can be True or False
                                            # ConstPhotRatio = True: the sum of convolution kernel is restricted to be a 
                                            #                constant across the field. 
                                            # ConstPhotRatio = False: the flux scaling between images is modeled by a 
                                            #                polynomial with degree -KerPolyOrder.

        # ----------------------------- Input & Output --------------------------------- #

        -PixA_REF_GPU []                    # Cupy ndarray for input reference image.

        -PixA_SCI_GPU []                    # Cupy ndarray for input science image.

        -PixA_mREF_GPU []                   # Cupy ndarray for input masked reference image (NaN-free).

        -PixA_mSCI_GPU []                   # Cupy ndarray for input masked science image (NaN-free).

        -PixA_DIFF_GPU [None]               # Cupy ndarray for output difference image.

        -PixA_Solution_GPU [None]           # Cupy ndarray for the solution of the linear system.
                                            # an array of (..., a_ijab, ... b_pq, ...).

        # ----------------------------- Miscellaneous --------------------------------- #
        
        -VERBOSE_LEVEL [2]                  # The level of verbosity, can be [0, 1, 2]
                                            # 0/1/2: QUIET/NORMAL/FULL mode

        # Important Notice:
        #
        # a): if reference is convolved in SFFT (-ForceConv='REF'), then DIFF = SCI - Convolved_REF.
        #     [the difference image is expected to have PSF & flux zero-point consistent with science image]
        #
        # b): if science is convolved in SFFT (-ForceConv='SCI'), then DIFF = Convolved_SCI - REF
        #     [the difference image is expected to have PSF & flux zero-point consistent with reference image]
        #
        # Remarks: this convention is to guarantee that transients emerge on science image always 
        #          show a positive signal on difference images.

        """

        # * assertions
        assert cp.sum(cp.isnan(PixA_mREF_GPU)) == 0, "The masked reference image contains NaNs!"
        assert cp.sum(cp.isnan(PixA_mSCI_GPU)) == 0, "The masked science image contains NaNs!"

        assert PixA_REF_GPU.ndim == 2, "The input PixA_REF_GPU is not two-dimensional!"
        assert PixA_SCI_GPU.ndim == 2, "The input PixA_SCI_GPU is not two-dimensional!"
        assert PixA_mREF_GPU.ndim == 2, "The input PixA_mREF_GPU is not two-dimensional!"
        assert PixA_mSCI_GPU.ndim == 2, "The input PixA_mSCI_GPU is not two-dimensional!"

        assert PixA_REF_GPU.dtype == cp.float64, "The array does not have dtype cp.float64!"
        assert PixA_SCI_GPU.dtype == cp.float64, "The array does not have dtype cp.float64!"
        assert PixA_mREF_GPU.dtype == cp.float64, "The array does not have dtype cp.float64!"
        assert PixA_mSCI_GPU.dtype == cp.float64, "The array does not have dtype cp.float64!"

        assert ForceConv in ['REF', 'SCI']
        ConvdSide = ForceConv
        KerHW = GKerHW

        device = cp.cuda.Device(int(CUDA_DEVICE_4SUBTRACT))
        device.use()
        
        # * Create the union NaN mask
        NaNmask_REF_GPU = cp.isnan(PixA_REF_GPU)
        NaNmask_SCI_GPU = cp.isnan(PixA_SCI_GPU)

        NaNmask_GPU = None
        if NaNmask_REF_GPU.any() or NaNmask_SCI_GPU.any():
            NaNmask_GPU = cp.logical_or(NaNmask_REF_GPU, NaNmask_SCI_GPU)

        # * Compile Functions in SFFT Subtraction
        if VERBOSE_LEVEL in [0, 1, 2]:
            print('MeLOn CheckPoint: TRIGGER Function Compilations of SFFT-SUBTRACTION!')

        Tcomp_start = time.time()
        NX, NY = PixA_REF_GPU.shape
        SFFTConfig = SingleSFFTConfigure.SSC(NX=NX, NY=NY, KerHW=KerHW, \
            KerPolyOrder=KerPolyOrder, BGPolyOrder=BGPolyOrder, ConstPhotRatio=ConstPhotRatio, \
            BACKEND_4SUBTRACT="Cupy", VERBOSE_LEVEL=VERBOSE_LEVEL)

        if VERBOSE_LEVEL in [1, 2]:
            _message = 'Function Compilations of SFFT-SUBTRACTION TAKES [%.3f s]' %(time.time() - Tcomp_start)
            print('\nMeLOn Report: %s' %_message)

        # * Perform SFFT Subtraction
        if ConvdSide == 'REF':
            PixA_mI_GPU, PixA_mJ_GPU = PixA_mREF_GPU, PixA_mSCI_GPU
            if NaNmask_GPU is not None:
                PixA_I_GPU, PixA_J_GPU = PixA_REF_GPU.copy(), PixA_SCI_GPU.copy()
                PixA_I_GPU[NaNmask_GPU] = PixA_mI_GPU[NaNmask_GPU]
                PixA_J_GPU[NaNmask_GPU] = PixA_mJ_GPU[NaNmask_GPU]
            else: PixA_I_GPU, PixA_J_GPU = PixA_REF_GPU, PixA_SCI_GPU

        if ConvdSide == 'SCI':
            PixA_mI_GPU, PixA_mJ_GPU = PixA_mSCI_GPU, PixA_mREF_GPU
            if NaNmask_GPU is not None:
                PixA_I_GPU, PixA_J_GPU = PixA_SCI_GPU.copy(), PixA_REF_GPU.copy()
                PixA_I_GPU[NaNmask_GPU] = PixA_mI_GPU[NaNmask_GPU]
                PixA_J_GPU[NaNmask_GPU] = PixA_mJ_GPU[NaNmask_GPU]
            else: PixA_I_GPU, PixA_J_GPU = PixA_SCI_GPU, PixA_REF_GPU

        Tsub_start = time.time()
        Solution_GPU, PixA_DIFF_GPU, _ = GeneralSFFTSubtract_PureCupy.GSS(
            PixA_I_GPU=PixA_I_GPU, PixA_J_GPU=PixA_J_GPU, PixA_mI_GPU=PixA_mI_GPU, PixA_mJ_GPU=PixA_mJ_GPU, \
            SFFTConfig=SFFTConfig, ContamMask_I_GPU=None, VERBOSE_LEVEL=VERBOSE_LEVEL
        )
        if VERBOSE_LEVEL in [1, 2]:
            _message = 'SFFT-SUBTRACTION TAKES [%.3f s]' %(time.time() - Tsub_start)
            print('\nMeLOn Report: %s' %_message)

        # * Modifications on the difference image
        #   a) when REF is convolved, DIFF = SCI - Conv(REF)
        #      PSF(DIFF) is coincident with PSF(SCI), transients on SCI are positive signal in DIFF.
        #   b) when SCI is convolved, DIFF = Conv(SCI) - REF
        #      PSF(DIFF) is coincident with PSF(REF), transients on SCI are still positive signal in DIFF.

        if NaNmask_GPU is not None:
            # ** Mask Union-NaN region
            PixA_DIFF_GPU[NaNmask_GPU] = np.nan
        
        if ConvdSide == 'SCI': 
            # ** Flip difference when science is convolved
            PixA_DIFF_GPU *= -1.
            
        return Solution_GPU, PixA_DIFF_GPU



import numpy as np

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.0"

class SkyLevel_Estimator: 
    @staticmethod
    def SLE(PixA_obj):
        
        # D. Jones - 2/13/14 | from PythonPhot package
        # equally idl commend: idl -e "sky,readfits('rd.b160504.000147.fits')"
        """This code is from the IDL Astronomy Users Library"""
        def mmm(sky_vector, highbad=False, debug=False, readnoise=False, \
            nsky=False, integer="discrete", mxiter=50, minsky=20, nan=True):

            """Estimate the sky background in a stellar contaminated field.

            MMM assumes that contaminated sky pixel values overwhelmingly display 
            POSITIVE departures from the true value.  Adapted from DAOPHOT 
            routine of the same name.
            
            CALLING SEQUENCE:
                skymod,sigma,skew = mmm.mmm( sky, highbad= , readnoise=, debug=, 
                                            minsky=, nsky=, integer=)
            
            INPUTS:
                sky - Array or Vector containing sky values.  This version of
                        MMM does not require SKY to be sorted beforehand.
            
            RETURNS:
                skymod - Scalar giving estimated mode of the sky values
                sigma -  Scalar giving standard deviation of the peak in the sky
                        histogram.  If for some reason it is impossible to derive
                        skymod, then SIGMA = -1.0
                skew -   Scalar giving skewness of the peak in the sky histogram
            
                If no output variables are supplied or if "debug" is set
                then the values of skymod, sigma and skew will be printed.
            
            OPTIONAL KEYWORD INPUTS:
                highbad - scalar value of the (lowest) "bad" pixel level (e.g. cosmic 
                            rays or saturated pixels) If not supplied, then there is 
                            assumed to be no high bad pixels.
                minsky - Integer giving mininum number of sky values to be used.   MMM
                            will return an error if fewer sky elements are supplied.
                            Default = 20.
                maxiter - integer giving maximum number of iterations allowed,default=50
                readnoise - Scalar giving the read noise (or minimum noise for any 
                            pixel).  Normally, MMM determines the (robust) median by 
                            averaging the central 20% of the sky values.  In some cases
                            where the noise is low, and pixel values are quantized a
                            larger fraction may be needed.  By supplying the optional
                            read noise parameter, MMM is better able to adjust the
                            fraction of pixels used to determine the median.                
                integer - Set this keyword if the  input SKY vector only contains
                            discrete integer values.  This keyword is only needed if the
                            SKY vector is of type float or double precision, but contains 
                            only discrete integer values.  (Prior to July 2004, the
                            equivalent of /INTEGER was set for all data types)
                debug -   If this keyword is set and non-zero, then additional 
                            information is displayed at the terminal.

            OPTIONAL OUTPUT KEYWORD:
                nsky - Integer scalar giving the number of pixels actually used for the
                        sky computation (after outliers have been removed).

            NOTES:
                (1) Program assumes that low "bad" pixels (e.g. bad CCD columns) have
                    already been deleted from the SKY vector.
                (2) MMM was updated in June 2004 to better match more recent versions
                    of DAOPHOT.
                (3) Does not work well in the limit of low Poisson integer counts
                (4) MMM may fail for strongly skewed distributions.

            METHOD:
                The algorithm used by MMM consists of roughly two parts:
                (1) The average and sigma of the sky pixels is computed.  These values
                        are used to eliminate outliers, i.e. values with a low probability
                        given a Gaussian with specified average and sigma.  The average
                        and sigma are then recomputed and the process repeated up to 20
                        iterations.
                (2) The amount of contamination by stars is estimated by comparing the 
                        mean and median of the remaining sky pixels.  If the mean is larger
                        than the median then the true sky value is estimated by
                        3*median - 2*mean
                    
            REVISION HISTORY:
                Adapted to IDL from 1986 version of DAOPHOT in STSDAS        W. Landsman, STX           Feb,      1987
                Added HIGHBAD keyword                                        W. Landsman                January,  1991
                Fixed occasional problem with integer inputs                 W. Landsman                Feb,      1994
                Avoid possible 16 bit integer overflow                       W. Landsman                November, 2001
                Added READNOISE, NSKY keywords,  new median computation      W. Landsman                June,     2004
                Added INTEGER keyword                                        W. Landsman                July,     2004
                Improve numerical precision                                  W. Landsman                October,  2004
                Fewer aborts on strange input sky histograms                 W. Landsman                October,  2005
                Added /SILENT keyword                                                                   November, 2005
                Fix too many /CON keywords to MESSAGE                        W.L.                       December, 2005
                Fix bug introduced June 2004 removing outliers               N. Cunningham/W. Landsman  January,  2006
                    when READNOISE not set
                Make sure that MESSAGE never aborts                          W. Landsman                January,  2008
                Add mxiter keyword and change default to 50                  W. Landsman                August,   2011
                Added MINSKY keyword                                         W.L.                       December, 2011
                Converted to Python                                          D. Jones                   January,  2014
            """

            if nan: sky_vector = sky_vector[np.where(sky_vector == sky_vector)]
            nsky = len( sky_vector )            #Get number of sky elements 
        
            if nsky < minsky:
                sigma=-1.0 ;  skew = 0.0; skymod = np.nan
                print('ERROR -Input vector must contain at least '+str(minsky)+' elements')
                return(skymod,sigma,skew)
        
            nlast = nsky-1                        #Subscript of last pixel in SKY array
            if debug:
                print('Processing '+str(nsky) + ' element array')
            sz_sky = np.shape(sky_vector)

            sky = np.sort(sky_vector)    #Sort SKY in ascending values

            skymid = 0.5*sky[int((nsky-1)/2)] + 0.5*sky[int(nsky/2)]  #Median value of all sky values
            
            cut1 = np.min( [skymid-sky[0],sky[nsky-1] - skymid] ) 
            if highbad: 
                cut1[np.where(cut1 > highbad - skymid)[0]] = highbad - skymid
            cut2 = skymid + cut1
            cut1 = skymid - cut1

            # Select the pixels between Cut1 and Cut2

            good = np.where( (sky <= cut2) & (sky >= cut1))[0]
            Ngood = len(good)

            if ( Ngood == 0 ):
                sigma=-1.0 ;  skew = 0.0; skymod = 0.0   
                print('ERROR - No sky values fall within ' + str(cut1) + \
            ' and ' + str(cut2))
                return(skymod,sigma,skew)
        
            delta = sky[good] - skymid  #Subtract median to improve arithmetic accuracy
            sum = np.sum(delta.astype('float64'))
            sumsq = np.sum(delta.astype('float64')**2)

            maximm = np.max( good) ; minimm = np.min(good)  # Highest value accepted at upper end of vector
            minimm = minimm -1               #Highest value reject at lower end of vector

            # Compute mean and sigma (from the first pass).

            skymed = 0.5*sky[int((minimm+maximm+1)/2)] + 0.5*sky[int((minimm+maximm)/2 + 1)] #median 
            skymn = sum/(maximm-minimm)                            #mean       
            sigma = np.sqrt(sumsq/(maximm-minimm)-skymn**2)             #sigma          
            skymn = skymn + skymid         #Add median which was subtracted off earlier 

            #    If mean is less than the mode, then the contamination is slight, and the
            #    mean value is what we really want.
            #    skymod =  (skymed < skymn) ? 3.*skymed - 2.*skymn : skymn
            if skymed < skymn:
                skymod = 3.*skymed - 2.*skymn
            else: skymod = skymn

            # Rejection and recomputation loop:

            niter = 0
            clamp = 1
            old = 0
            # START_LOOP:
            redo = True
            while redo:
                niter = niter + 1                     
                if ( niter > mxiter ):
                    sigma=-1.0 ;  skew = 0.0   
                    print('ERROR - Too many ('+str(mxiter) + ') iterations,' + \
                            ' unable to compute sky')
                    #            import pdb; pdb.set_trace()
                    return(skymod,sigma,skew)

                if ( maximm-minimm < minsky ):    #Error? 

                    sigma = -1.0 ;  skew = 0.0   
                    print('ERROR - Too few ('+str(maximm-minimm) +  \
                            ') valid sky elements, unable to compute sky')
                    return(skymod,sigma,skew)

                # Compute Chauvenet rejection criterion.

                r = np.log10( float( maximm-minimm ) )      
                r = np.max( [ 2., ( -0.1042*r + 1.1695)*r + 0.8895 ] )

                # Compute rejection limits (symmetric about the current mode).

                cut = r*sigma + 0.5*np.abs(skymn-skymod)   
                #    if integer: cut = cut > 1.5 
                cut1 = skymod - cut   ;    cut2 = skymod + cut

                # 
                # Recompute mean and sigma by adding and/or subtracting sky values
                # at both ends of the interval of acceptable values.
            
                redo = False
                newmin = minimm             
                if sky[newmin+1] >= cut1: tst_min = 1      #Is minimm+1 above current CUT?
                else: tst_min = 0
                if (newmin == -1) and tst_min: done = 1    #Are we at first pixel of SKY?
                else: done = 0
                if not done:
                    if newmin > 0: skyind = newmin
                    else: skyind = 0
                    if (sky[skyind] < cut1) and tst_min: done = 1
                if not done:
                    istep = 1 - 2*int(tst_min)
                    while not done:
                        newmin = newmin + istep
                        if (newmin == -1) | (newmin == nlast): done = 1
                        if not done:
                            if (sky[newmin] <= cut1) and (sky[newmin+1] >= cut1): done = 1
                
                    if tst_min:  delta = sky[newmin+1:minimm+1] - skymid
                    else: delta = sky[minimm+1:newmin+1] - skymid
                    sum = sum - istep*np.sum(delta)
                    sumsq = sumsq - istep*np.sum(delta**2)
                    redo = True
                    minimm = newmin

                newmax = maximm
                if sky[maximm] <= cut2: tst_max = 1           #Is current maximum below upper cut?
                else: tst_max = 0
                if (maximm == nlast) and tst_max: done = 1
                else: done = 0                  #Are we at last pixel of SKY array?
                if not done:
                    if maximm+1 < nlast: skyind = maximm+1
                    else: skyind = nlast
                    if ( tst_max ) and (sky[skyind] > cut2): done = 1 
                if not done: # keep incrementing newmax
                    istep = -1 + 2*int(tst_max)         #Increment up or down?
                    while not done:
                        newmax = newmax + istep
                        if (newmax == nlast) or (newmax == -1): done = 1
                        if not done:
                            if ( sky[newmax] <= cut2 ) and ( sky[newmax+1] >= cut2 ): done = 1

                    if tst_max: 
                        delta = sky[maximm+1:newmax+1] - skymid
                    else:
                        delta = sky[newmax+1:maximm+1] - skymid
                    sum = sum + istep*np.sum(delta)
                    sumsq = sumsq + istep*np.sum(delta**2)
                    redo = True
                    maximm = newmax

                #       
                # Compute mean and sigma (from this pass).
                #
                nsky = maximm - minimm
                if ( nsky < minsky ): # error?
                    sigma = -1.0 ;  skew = 0.0   
                    print('ERROR - Outlier rejection left too few sky elements')
                    return(skymod,sigma,skew)

                skymn = sum/nsky
                var = sumsq/nsky - skymn**2
                if var < 0: var = 0
                sigma = float( np.sqrt( var ))
                skymn = skymn + skymid 

                #  Determine a more robust median by averaging the central 20% of pixels.
                #  Estimate the median using the mean of the central 20 percent of sky
                #  values.   Be careful to include a perfectly symmetric sample of pixels about
                #  the median, whether the total number is even or odd within the acceptance
                #  interval
            
                center = (minimm + 1 + maximm)/2.
                side = np.round(0.2*(maximm-minimm))/2.  + 0.25
                j = np.round(center-side)
                k = np.round(center+side)

                #  In case  the data has a large number of of the same (quantized) 
                #  intensity, expand the range until both limiting values differ from the 
                #  central value by at least 0.25 times the read noise.

                if readnoise:
                    L = round(center-0.25)
                    M = round(center+0.25)
                    R = 0.25*readnoise
                    while ((j > 0) and (k < nsky-1) and \
                            ( ((sky[int(L)] - sky[int(j)]) < R) or ((sky[int(k)] - sky[int(M)]) < R))):
                        j -= 1
                        k += 1

                skymed = np.sum(sky[int(j):int(k+1)])/(k-j+1)

                #  If the mean is less than the median, then the problem of contamination
                #  is slight, and the mean is what we really want.

                if skymed < skymn : 
                    dmod = 3.*skymed-2.*skymn-skymod 
                else: dmod = skymn - skymod

                # prevent oscillations by clamping down if sky adjustments are changing sign
                if dmod*old < 0: clamp = 0.5*clamp
                skymod = skymod + clamp*dmod 
                old = dmod     

            #   if redo then goto, START_LOOP     
            skew = float( (skymn-skymod)/max([1.,sigma]) )
            nsky = maximm - minimm 

            if debug:
                print('% MMM: Number of unrejected sky elements: ', str(nsky,2), \
                        '    Number of iterations: ',  str(niter))
                print('% MMM: Mode, Sigma, Skew of sky vector:', skymod, sigma, skew   )

            return(skymod,sigma,skew)

        BGLevel, BGSig = mmm(sky_vector=PixA_obj)[:2]
        
        """
        # > Alternative Way to use mmm algorithm through photutils
        from astropy.stats import SigmaClip
        from photutils import MMMBackground, StdBackgroundRMS

        sigma_clip = SigmaClip(sigma=3.0, maxiters=5)
        bkg = MMMBackground(sigma_clip=sigma_clip)
        BGLevel = bkg.calc_background(PixA_obj)
        # NOTE: Here calculate the background RMS in an array as the (sigma-clipped) standard deviation
        bkgrms = StdBackgroundRMS(sigma_clip=sigma_clip)
        BGSig = bkgrms.calc_background_rms(PixA_obj)    

        """

        return BGLevel, BGSig



import math
import threading
import multiprocessing
# version: Feb 6, 2023

__author__ = "Lei Hu <leihu@andrew.cmu.edu>"
__version__ = "v1.4"

class Multi_Proc:
    @staticmethod
    def MP(taskid_lst=None, func=None, nproc=8, mode='mp'):
        
        if mode == 'mp':
            # Ref: https://gist.github.com/tappoz/cb88f7a9d9ba27cfee1e2f03537fac16
            
            def worker(taskid_lst, out_q):
                outdict = {}
                for tid in taskid_lst:
                    outdict[tid] = func(tid)
                out_q.put(outdict)
            
            out_q = multiprocessing.Queue()
            chunksize = int(math.ceil(len(taskid_lst) / float(nproc)))
            
            procs = []
            for i in range(nproc):
                p = multiprocessing.Process(target=worker, \
                    args=(taskid_lst[chunksize * i:chunksize * (i + 1)], out_q)) 
                procs.append(p)
                p.start()

            resultdict = {}
            for i in range(nproc):
                resultdict.update(out_q.get())

            for p in procs:
                p.join()

            return resultdict
        
        if mode == 'threading':
            # NOTE: This mode cannot return a result dictionary
            def trigger(klst):
                for k in klst: func(k)
                return None

            myThread_Queue = []
            chunksize = int(math.ceil(len(taskid_lst) / float(nproc)))
            for i in range(nproc):
                taskid_asslst = taskid_lst[chunksize * i:chunksize * (i + 1)]
                myThread = threading.Thread(target=trigger, args=(taskid_asslst,))
                myThread_Queue.append(myThread)
                myThread.start()
            
            for t in myThread_Queue:
                t.join()
            
            return None
